from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def radar_chart(
    self: "PlotLayer",
    order: int,
    names: str,
    data: Union[str, DataFrame, list[dict]],
    values: Optional[list[str]] = None,
    max_field: Optional[str] = None,
    fill_area: bool = False,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a radar chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    series_options["type"] = "radar"
    del common_options["yAxis"]
    del common_options["xAxis"]
    del common_options["grid"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["radar"] = {"indicator": "#set_data#"}
    if fill_area:
        series_options["areaStyle"] = {"opacity": 0.25}

    data_mapping_to_tuples = await self._choose_data(order, data)
    if not values:
        values = [
            f for f in data_mapping_to_tuples.keys() if f != names and f != max_field
        ]

    series_options["data"] = [{"value": "#set_data#", "name": vf} for vf in values]
    common_options["series"] = [series_options]
    indicator_fields = (
        {"name": names} if not max_field else {"name": names, "max": max_field}
    )
    await self._create_echart(
        fields=[indicator_fields, *values],
        options=common_options,
        data_mapping_to_tuples=data_mapping_to_tuples,
        option_modifications=option_modifications,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



# This file was generated automatically by scripts/user_classes_header_generator.py do not modify it!
# If the user access files are modified, this file has to be regenerated with the script.
from typing import Optional, Union
from pandas import DataFrame
from dataclasses import dataclass
import shimoku


class ActionsLayerHeader:
    """
    This class is used to interact with the API at the Workspace layer
    """

    def create_action(
        self,
        name: str,
        description: Optional[str] = None,
        action_template_id: Optional[str] = None,
        universe_api_key: Optional[str] = None,
        path_to_code: Optional[str] = None,
        code: Optional[str] = None,
        overwrite: Optional[bool] = False,
        libraries: Optional[list[str]] = None,
    ) -> str:
        """
        Create an action.
        :param name: name of the action
        :param description: description of the action
        :param action_template_id: id of the action template
        :param universe_api_key: id of the universe API key
        :param path_to_code: path to the code
        :param code: code of the action
        :param libraries: libraries of the action
        :param overwrite: whether to overwrite the action if it already exists
        """
        pass

    def delete_action(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ):
        """
        Delete an action.
        :param uuid: uuid of the action
        :param name: name of the action
        """
        pass

    def get_action_code(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> str:
        """
        Get the code of the action.
        :param uuid: uuid of the action
        :param name: name of the action
        :return: The code of the action
        """
        pass

    def get_action_metadata(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
    ) -> dict:
        """
        Get an action.
        :param uuid: uuid of the action
        :param name: name of the action
        :return: The action
        """
        pass

    def update_action(
        self,
        uuid: Optional[str] = None,
        name: Optional[str] = None,
        new_name: Optional[str] = None,
        new_description: Optional[str] = None,
    ):
        """
        Update an action.
        :param uuid: uuid of the action
        :param name: name of the action
        :param new_name: new name of the action
        :param new_description: new description of the action
        """
        pass



from shimoku import Client


def iframe(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "iframe-test")
    url = "https://www.marca.com/"
    shimoku_client.plt.iframe(url=url, order=0)
    shimoku_client.plt.iframe(
        url=url, order=1, height=160 * 8, cols_size=6, padding="0,3,0,3"
    )



from shimoku import Client
from .data.bentobox_indicator import bentobox_indicator_data


def bentobox(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "bentobox-test")

    shimoku_client.plt.set_bentobox(cols_size=8, rows_size=3)

    shimoku_client.plt.indicator(
        data=bentobox_indicator_data, order=0, rows_size=10, cols_size=12
    )
    shimoku_client.plt.indicator(
        data=bentobox_indicator_data, order=1, rows_size=10, cols_size=12
    )
    shimoku_client.plt.bar(
        data="main data", x="date", y=["x", "y"], order=2, rows_size=26, cols_size=24
    )

    shimoku_client.plt.pop_out_of_bentobox()



import requests
import pandas as pd

res = requests.get(
    url="https://echarts.apache.org/examples/data/asset/data/aqi-beijing.json"
)

# Convert it to json
data = res.json()[-100:]

# Remap data
segmented_line_data = pd.DataFrame(
    [
        {
            "date": data[i][0],
            "y": data[i][1],
            "y_displaced": data[(i + 10) % len(data)][1],
            "y_multiplied": data[i][1] * 2,
        }
        for i in range(len(data))
    ]
)



"""
Used as base Mailchimp:
https://github.com/mailchimp/mailchimp-marketing-python/blob/master/mailchimp_marketing/api_client.py
"""

from typing import Optional, Callable

import datetime
import asyncio

from shimoku.exceptions import APIError
from shimoku.utils import IN_BROWSER
import json
from pkg_resources import get_distribution


import logging
from shimoku.execution_logger import ClassWithLogging, log_error

logger = logging.getLogger(__name__)

SHIMOKU_VERSION_KEY = "python-shimoku-pkg-version"


def get_request_function():
    """Auxiliary function to get the appropriate request function"""
    if IN_BROWSER:
        from pyodide.http import pyfetch

        async def request(
            self: "ApiClient",
            method,
            url,
            query_params,
            headers,
            body,
        ):
            params = dict(
                method=method,
                headers=headers,
                body=body
                if isinstance(body, bytes)
                else json.dumps(body).encode("utf-8"),
                timeout=self.timeout,
                params=query_params,
            )
            if not body:
                del params["body"]
            res = await pyfetch(url, **params)
            if (
                hasattr(res, "headers")
                and "application/json" in res.headers.get("content-type")
            ) or ((hasattr(res, "json")) and not hasattr(res, "read")):
                data = await res.json()
            else:
                data = await res.read()
            if not res.ok:
                self.raise_api_exception(data)
            return data

    else:
        import aiohttp

        async def request(
            self: "ApiClient",
            method,
            url,
            query_params,
            headers,
            body,
        ):
            session_params = {
                "auth": aiohttp.BasicAuth("user", self.api_key)
                if self.is_basic_auth
                else None
            }
            if not session_params["auth"]:
                del session_params["auth"]

            request_options = {
                "method": method,
                "url": url,
                "headers": headers,
                "params": query_params,
                "json": body,
            }
            if not headers:
                del request_options["headers"]
            if not query_params:
                del request_options["params"]
            if not body:
                del request_options["json"]
            elif isinstance(body, bytes):
                request_options["data"] = body
                del request_options["json"]

            async with aiohttp.ClientSession(**session_params) as session:
                async with session.request(**request_options) as res:
                    if (
                        "content-type" in res.headers
                        and "application/json" in res.headers.get("content-type")
                    ):
                        data = await res.json()
                    else:
                        data = await res.read()
                    if not res.ok:
                        self.raise_api_exception(data)
                    return data

    return request


def retry(a_func: Callable, n_retries: int):
    """Auxiliary function to retry a function n_retries times waiting exponentially for each retry"""

    async def wrapper(*args, **kwargs):
        for i in range(n_retries):
            try:
                return await a_func(*args, **kwargs)
            except APIError as e:
                await asyncio.sleep(2**i)
                if i == n_retries - 1:
                    raise e
                logger.warning(f"Error: {e}")

    return wrapper


class ApiClient(ClassWithLogging):
    PRIMITIVE_TYPES = (float, int, bool, bytes, str)

    _module_logger = logger

    def __init__(
        self,
        environment: str,
        playground: bool,
        config=None,
        server_host="127.0.0.1",
        server_port=8000,
        retry_attempts: int = 5,
    ):
        self.cache_enabled = True
        self.environment = environment
        self.playground = playground
        self.retry_attempts = retry_attempts

        if config is None:
            config = {}

        if "production".startswith(environment.lower()):
            self.host = "https://api.shimoku.io/external/v1/"
        elif "develop".startswith(environment.lower()):
            self.host = "https://api.develop.shimoku.io/external/v1/"
        elif environment == "guillermo":
            self.host = "https://wxauh7u2te.execute-api.eu-west-1.amazonaws.com/guillermo/external/v1/"
        else:
            raise ValueError(
                f'The namespace must be either "production" or "develop | '
                f"namespace introduced: {environment}"
            )
        if playground:
            self.host = f"http://{server_host}:{server_port}/external/v1/"

        # semaphor for async api calls
        self.semaphore_limit = 10
        self.semaphore = None

        # DEFAULTS
        # Api key
        self.api_key: str = ""
        self.is_basic_auth: bool = False
        # OAuth
        self.access_token: str = ""
        self.is_oauth: bool = False
        # General
        self.server: str = "invalid-server"
        self.timeout: int = 120

        self.default_headers = {
            "Content-Type": "application/json",
        }
        self.set_config(config)
        self.call_counter = 0
        # Default vars

    def set_config(self, config={}):
        """Set all config values"""
        # Basic Auth
        self.api_key: str = config["api_key"] if "api_key" in config.keys() else ""
        self.is_basic_auth: bool = self.api_key != ""

        # OAuth
        self.access_token: str = (
            config["access_token"] if "access_token" in config.keys() else ""
        )
        self.is_oauth: bool = self.access_token != ""

        if not self.is_oauth and not self.is_basic_auth:
            raise ValueError("You must provide either an API Key or Access Token")

        # If using Basic auth and no server is provided,
        # attempt to extract it from the api_key directly.
        self.server: str = (
            config["server"] if "server" in config.keys() else "invalid-server"
        )
        if self.server == "invalid-server" and self.is_basic_auth:
            self.server: str = self.get_server_from_api_key(self.api_key)

        self.timeout = config["timeout"] if "timeout" in config.keys() else 120

    async def call_api(
        self,
        resource_path,
        method,
        path_params=None,
        query_params=None,
        header_params=None,
        body=None,
        collection_formats=None,
        limit: Optional[int] = None,
        elastic_supported: bool = False,
        **kwargs,
    ):
        """Create and call the API request with headers, params and others"""
        # header parameters
        header_params = header_params or {}
        header_params.update(self.default_headers)
        if header_params:
            header_params = self.sanitize_for_serialization(header_params)
            header_params = dict(
                self.parameters_to_tuples(header_params, collection_formats)
            )

        # path parameters
        if path_params:
            path_params = self.sanitize_for_serialization(path_params)
            path_params = self.parameters_to_tuples(path_params, collection_formats)
            for k, v in path_params:
                # specified safe chars, encode everything
                resource_path = resource_path.replace("{%s}" % k, str(v))

        # query parameters
        if query_params:
            query_params = self.sanitize_for_serialization(query_params)
            query_params = self.parameters_to_tuples(query_params, collection_formats)

        # request url
        url = self.host + resource_path

        if self.server:
            url = url.replace("server", self.server)

        if not self.semaphore:
            self.semaphore = asyncio.Semaphore(self.semaphore_limit)

        try:
            # perform request and return response
            async with self.semaphore:
                return await self.request(
                    method,
                    url,
                    query_params,
                    headers=header_params,
                    body=body,
                    limit=limit,
                    elastic_supported=elastic_supported,
                )
        except asyncio.CancelledError:
            self.semaphore = asyncio.Semaphore(self.semaphore_limit)
            async with self.semaphore:
                return await self.request(
                    method,
                    url,
                    query_params,
                    headers=header_params,
                    body=body,
                    limit=limit,
                    elastic_supported=elastic_supported,
                )

    def set_http_info(self, **kwargs):  # noqa: E501
        """
        This method makes a synchronous HTTP request by default. To make an
        asynchronous HTTP request, please pass async_req=True
        >>> thread = api.remove_with_http_info(campaign_id, async_req=True)
        >>> result = thread.get()
        :param async_req bool
        :param str campaign_id: The unique id for the campaign. (required)
        :return: None
                 If the method is called asynchronously,
                 returns the request thread.
        """
        params = locals()
        for key, val in params["kwargs"].items():
            params[key] = val
        body_params = params["kwargs"].get("data")
        del params["kwargs"]

        collection_formats = {}

        query_params = []

        header_params = {}

        form_params = []
        local_var_files = {}

        # HTTP header `Accept`
        header_params["Accept"] = self.select_header_accept(
            ["application/json", "application/problem+json"]
        )  # noqa: E501

        # HTTP header `Content-Type`
        header_params["Content-Type"] = self.select_header_content_type(  # noqa: E501
            ["application/json"]
        )  # noqa: E501

        # Authentication setting
        auth_settings = ["basicAuth"]  # noqa: E501

        return (
            query_params,
            header_params,
            kwargs.get("body_params"),
            form_params,
            local_var_files,
            auth_settings,
            params,
            collection_formats,
        )

    async def query_element(
        self,
        method: str,
        endpoint: str,
        limit: Optional[int] = None,
        elastic_supported: bool = False,
        **kwargs,
    ) -> dict:
        """Retrieve an element if the endpoint exists

        :param method: examples are 'GET', 'POST', etc
        :param endpoint: example: 'business/{businessId}/app/{appId}
        :param limit: limit the number of results returned
        :param elastic_supported: whether the endpoint supports elastic search
        """
        (
            query_params,
            header_params,
            body_params,
            form_params,
            local_var_files,
            auth_settings,
            params,
            collection_formats,
        ) = self.set_http_info(**kwargs)

        path_params = {}
        if endpoint in params:
            path_params[endpoint] = params[endpoint]  # noqa: E501

        element_data: dict = await retry(self.call_api, n_retries=self.retry_attempts)(
            endpoint,
            method,
            path_params,
            query_params,
            header_params,
            limit=limit,
            elastic_supported=elastic_supported,
            body=body_params,
            post_params=form_params,
            files=local_var_files,
            response_type=None,  # noqa: E501
            auth_settings=auth_settings,
            async_req=params.get("async_req"),
            _return_http_data_only=params.get("_return_http_data_only"),
            _preload_content=params.get("_preload_content", True),
            _request_timeout=params.get("_request_timeout"),
            collection_formats=collection_formats,
        )

        if kwargs.get("progress_bar"):
            progress_bar, how_much = kwargs.get("progress_bar")
            if progress_bar:
                progress_bar.update(how_much)

        return element_data

    @staticmethod
    def raise_api_exception(response: str) -> None:
        """Raise an ApiClientError with the message changed to be more user friendly
        :param response: the response from the API
        """
        replace_words = {
            "report": "component",
            "app": "menu path",
            "business": "workspace",
            "dashboard": "board",
        }
        for word, replacement in replace_words.items():
            response = (
                response.replace(word, replacement)
                if isinstance(response, str)
                else response
            )
        log_error(logger, response, APIError)

    _request = get_request_function()

    async def request(
        self,
        method,
        url,
        query_params=None,
        headers=None,
        body=None,
        limit: Optional[int] = None,
        elastic_supported: bool = False,
        to_tazawa: bool = True,
    ):
        headers = headers or {}
        if to_tazawa:
            headers.update({"Authorization": "Bearer " + self.access_token})
            # TODO see how to handle from the browser
            # headers.update(
            #     {
            #         SHIMOKU_VERSION_KEY: get_distribution("shimoku").version
            #         if not IN_BROWSER
            #         else "2.0.0"
            #     }
            # )

        if method not in ["GET", "HEAD", "OPTIONS", "DELETE", "POST", "PUT", "PATCH"]:
            raise ValueError(
                "http method must be `GET`, `HEAD`, `OPTIONS`,"
                " `POST`, `PATCH`, `PUT` or `DELETE`."
            )
        body_from = 0
        next_token = None
        data_res = {} if not elastic_supported else []
        req_limit = limit if limit else 100
        method = method if not elastic_supported else "POST"

        while True:  # loop until nextToken is None
            aux_url = url
            if to_tazawa:
                if elastic_supported:
                    body = {"from": body_from, "limit": req_limit}
                elif method == "GET":
                    aux_url += (
                        f"?nextToken={next_token}"
                        if next_token
                        else f"?limit={req_limit}"
                    )

            logger.debug(
                f"method:{method}, url: {aux_url}, headers: {headers},"
                f"query params: {query_params}, body: {body}"
            )

            data = await self._request(
                method=method,
                url=aux_url,
                query_params=query_params,
                headers=headers,
                body=body,
            )
            self.call_counter += 1
            logger.debug(data)

            if not to_tazawa:
                return data

            if elastic_supported:
                data_res.extend(data)
                body_from += req_limit
                if limit:
                    limit -= req_limit
                if len(data) < req_limit or (limit and limit <= 0):
                    break
            elif data and "items" in data:
                if data_res.get("items"):
                    data_res["items"].extend(data.get("items"))
                else:
                    data_res = data
                next_token = data.get("nextToken")
                if not next_token:
                    break
            else:
                data_res = data
                break

        return data_res

    def sanitize_for_serialization(self, obj):
        """Builds a JSON POST object.
        If obj is None, return None.
        If obj is str, int, long, float, bool, return directly.
        If obj is datetime.datetime, datetime.date
            convert to string in iso8601 format.
        If obj is list, sanitize each element in the list.
        If obj is dict, return the dict.
        If obj is swagger model, return the properties dict.
        :param obj: The data to serialize.
        :return: The serialized form of data.
        """
        if obj is None:
            return None
        elif isinstance(obj, self.PRIMITIVE_TYPES):
            return obj
        elif isinstance(obj, list):
            return [self.sanitize_for_serialization(sub_obj) for sub_obj in obj]
        elif isinstance(obj, tuple):
            return tuple(self.sanitize_for_serialization(sub_obj) for sub_obj in obj)
        elif isinstance(obj, (datetime.datetime, datetime.date)):
            return obj.isoformat()

        if isinstance(obj, dict):
            obj_dict = obj
        else:
            # Convert model obj to dict except
            # attributes `swagger_types`, `attribute_map`
            # and attributes which value is not None.
            # Convert attribute name to json key in
            # model definition for request.
            obj_dict = {
                obj.attribute_map[attr]: getattr(obj, attr)
                for attr, _ in obj.items()
                if getattr(obj, attr) is not None
            }

        return {
            key: self.sanitize_for_serialization(val) for key, val in obj_dict.items()
        }

    @staticmethod
    def parameters_to_tuples(params, collection_formats):
        """Get parameters as list of tuples, formatting collections.
        :param params: Parameters as dict or list of two-tuples
        :param dict collection_formats: Parameter collection formats
        :return: Parameters as list of tuples, collections formatted
        """
        new_params = []
        if collection_formats is None:
            collection_formats = {}
        for k, v in (
            params.items() if isinstance(params, dict) else params
        ):  # noqa: E501
            if k in collection_formats:
                collection_format = collection_formats[k]
                if collection_format == "multi":
                    new_params.extend((k, value) for value in v)
                else:
                    if collection_format == "ssv":
                        delimiter = " "
                    elif collection_format == "tsv":
                        delimiter = "\t"
                    elif collection_format == "pipes":
                        delimiter = "|"
                    else:  # csv is the default
                        delimiter = ","
                    new_params.append((k, delimiter.join(str(value) for value in v)))
            else:
                new_params.append((k, v))
        return new_params

    @staticmethod
    def select_header_accept(accepts):
        """Returns `Accept` based on an array of accepts provided.
        :param accepts: List of headers.
        :return: Accept (e.g. application/json).
        """
        if not accepts:
            return

        accepts = [x.lower() for x in accepts]

        if "application/json" in accepts:
            return "application/json"
        else:
            return ", ".join(accepts)

    @staticmethod
    def get_server_from_api_key(api_key: str) -> str:
        try:
            split: list[str] = api_key.split("-")
            if len(split) == 2:
                return split[1]
            else:
                return "invalid-server"
        except:
            return ""

    @staticmethod
    def select_header_content_type(content_types):
        """Returns `Content-Type` based on an array of content_types provided.
        :param content_types: List of content-types.
        :return: Content-Type (e.g. application/json).
        """
        if not content_types:
            return "application/json"

        content_types = [x.lower() for x in content_types]

        if "application/json" in content_types or "*/*" in content_types:
            return "application/json"
        else:
            return content_types[0]



import pandas as pd
from os import path
from shimoku import Client
from .data.pie import pie_data

data_path = path.join(path.dirname(path.abspath(__file__)), "data")


def rose(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "rose")
    shimoku_client.plt.rose(data="main data", names="date", values="x", order=0)
    shimoku_client.plt.rose(data=pie_data, names="name", values="value", order=1)

    df = pd.read_csv(path.join(data_path, "test_stack_distribution.csv"))

    value_columns = [col for col in df.columns if col != "Segment"]
    df = df[["Segment"] + value_columns]

    rose_data = pd.DataFrame(columns=["name", "value"])
    df_transposed = df.transpose().reset_index().drop(0)
    value_columns = [col for col in df_transposed.columns if col != "index"]
    rose_data["value"] = df_transposed[value_columns].apply(
        lambda row: sum(row), axis=1
    )
    rose_data["name"] = df_transposed["index"]
    shimoku_client.plt.rose(
        data=rose_data, values="value", names="name", order=2, rows_size=3, cols_size=6
    )



from shimoku.cli.configuration import add_configuration_parser
from shimoku.cli.cloud_parser import add_cloud_parser
from shimoku.cli.playground import add_playground_parser
from shimoku.cli.persistence import add_persistence_parser
from shimoku.cli.listen import add_listener_parser
from shimoku.cli import CLIParser, CLIFuncParam

from shimoku.execution_logger import configure_logging

import asyncio

main_parser = CLIParser(
    arguments=[
        CLIFuncParam(
            name="interactive",
            arg_help="Run in interactive mode",
            mandatory=False,
            default=False,
            action="store_true",
            alt_name="i",
        ),
        CLIFuncParam(
            name="shell-commands-enabled",
            arg_help="Enable the possibility to run shell commands from the interactive mode",
            mandatory=False,
            default=False,
            action="store_true",
            alt_name="s",
        ),
    ]
)

add_configuration_parser(main_parser)
add_playground_parser(main_parser)
add_cloud_parser(main_parser)
add_persistence_parser(main_parser)
add_listener_parser(main_parser)


def main():
    configure_logging("INFO")
    try:
        asyncio.run(main_parser.parse_args())
    except KeyboardInterrupt:
        print("Interrupted by user")
    except asyncio.CancelledError:
        print("Interrupted by user")


if __name__ == "__main__":
    main()



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.exceptions import DataError

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


async def waterfall_chart(
    self: "PlotLayer",
    data: Union[str, DataFrame, list[dict]],
    order: int,
    x: str,
    positive: str,
    negative: str,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    title: Optional[str] = None,
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    show_balance: bool = False,
    variant: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    if isinstance(data, str):
        log_error(
            logger,
            "Waterfall cannot use shared data, please provide a DataFrame or a list of dicts",
            DataError,
        )

    df: DataFrame = validate_data_is_pandarable(data)

    df["Base"] = (df[positive].cumsum() - df[negative].cumsum()).shift(1).fillna(0)
    df["Value"] = df[positive] - df[negative]
    df["Balance"] = df["Base"] + df["Value"]
    df["VisibleNegBase"] = 0
    df["VisiblePosBase"] = 0

    df["Positive"] = df["Value"].apply(lambda x: x if x > 0 else 0)
    df["Negative"] = df["Value"].apply(lambda x: abs(x) if x < 0 else 0)

    def change_sign(row):
        if row["Base"] < 0:
            if row["Base"] + row["Positive"] > 0:
                row["VisiblePosBase"] = -(
                    row["Positive"] - (row["Base"] + row["Positive"])
                )
                row["Positive"] = row["Base"] + row["Positive"]
                row["Base"] = 0
                return row
            else:
                row["Positive"] = -row["Positive"]
                row["Negative"] = -row["Negative"]
        else:
            if row["Base"] - row["Negative"] < 0:
                row["VisibleNegBase"] = row["Negative"] + (
                    row["Base"] - row["Negative"]
                )
                row["Negative"] = row["Base"] - row["Negative"]
                row["Base"] = 0
                return row

        row["Base"] = (
            row["Base"] - row["Positive"]
            if row["Base"] < 0
            else row["Base"] - row["Negative"]
        )
        return row

    df = df.apply(change_sign, axis=1)
    df = df[
        [
            x,
            "Balance",
            "Base",
            "VisibleNegBase",
            "VisiblePosBase",
            "Negative",
            "Positive",
        ]
    ].fillna(0)

    data_mappings_to_tuples = await self._choose_data(order, df)

    common_options = get_common_echart_options()
    del common_options["toolbox"]["feature"]["magicType"]
    common_options.update(
        {
            "legend": {
                "data": [positive, negative] + (["Balance"] if show_balance else []),
                "show": True,
                "type": "scroll",
                "itemGap": 16,
                "icon": "circle",
            },
            "xAxis": [
                {
                    "data": "#set_data#",
                    "type": "category",
                    "fontFamily": "Rubik",
                    "nameLocation": "middle",
                    "nameGap": 35,
                }
            ],
            "yAxis": [
                {
                    "type": "value",
                    "fontFamily": "Rubik",
                    "nameLocation": "middle",
                    "nameGap": int(
                        24 + 7 * (df["Balance"].astype(str).apply(len).max())
                    ),
                },
            ],
        }
    )
    series_options = [
        {
            "data": "#set_data#",
            "name": "Balance",
            "step": "end",
            "symbol": "none",
            "type": "line",
            "itemStyle": {
                "color": ShimokuPalette.CHART_C3.value
                if show_balance
                else "transparent",
            },
            "emphasis": {
                "itemStyle": {
                    "color": ShimokuPalette.CHART_C3.value
                    if show_balance
                    else "transparent",
                },
                "lineStyle": {
                    "color": ShimokuPalette.CHART_C3.value
                    if show_balance
                    else "transparent",
                },
            },
            "zlevel": 0,
            "tooltip": {"show": show_balance},
        },
        {
            "data": "#set_data#",
            "name": "Placeholder",
            "type": "bar",
            "stack": "Total",
            "itemStyle": {"borderColor": "transparent", "color": "transparent"},
            "emphasis": {
                "itemStyle": {"borderColor": "transparent", "color": "transparent"}
            },
            "tooltip": {"show": False},
        },
        {
            "data": "#set_data#",
            "name": "Negative visible Base",
            "type": "bar",
            "stack": "Total",
            "silent": True,
            "itemStyle": {
                "color": ShimokuPalette.CHART_C6.value,
            },
            "emphasis": {
                "itemStyle": {
                    "color": ShimokuPalette.CHART_C6.value,
                },
            },
            "zlevel": 1,
            "tooltip": {"show": False},
        },
        {
            "data": "#set_data#",
            "name": "Positive visible Base",
            "type": "bar",
            "stack": "Total",
            "silent": True,
            "itemStyle": {
                "color": ShimokuPalette.CHART_C2.value,
            },
            "emphasis": {
                "itemStyle": {
                    "color": ShimokuPalette.CHART_C2.value,
                },
            },
            "zlevel": 1,
            "tooltip": {"show": False},
        },
        {
            "data": "#set_data#",
            "name": negative,
            "type": "bar",
            "stack": "Total",
            "itemStyle": {
                "borderRadius": [0, 0, 9, 9],
                "color": ShimokuPalette.CHART_C6.value,
            },
            "emphasis": {
                "itemStyle": {
                    "borderRadius": [0, 0, 9, 9],
                    "color": ShimokuPalette.CHART_C6.value,
                },
            },
            "zlevel": 1,
        },
        {
            "data": "#set_data#",
            "name": positive,
            "type": "bar",
            "stack": "Total",
            "itemStyle": {
                "borderRadius": [9, 9, 0, 0],
                "color": ShimokuPalette.CHART_C2.value,
            },
            "emphasis": {
                "itemStyle": {
                    "borderRadius": [9, 9, 0, 0],
                    "color": ShimokuPalette.CHART_C2.value,
                },
            },
            "zlevel": 1,
        },
    ]

    await self._create_trend_chart(
        data_mapping_to_tuples=data_mappings_to_tuples,
        order=order,
        axes=x,
        values=[
            "Balance",
            "Base",
            "VisibleNegBase",
            "VisiblePosBase",
            "Negative",
            "Positive",
        ],
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        title=title,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        variant=variant,
    )



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging

logger = logging.getLogger(__name__)


async def tree_chart(
    self: "PlotLayer",
    order: int,
    data: Union[str, dict, list[dict]],
    radial: bool = False,
    vertical: bool = False,
    title: Optional[str] = None,
    padding: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a tree chart"""
    if radial and vertical:
        logger.warning(
            "Radial and vertical options are mutually exclusive. Vertical option will be ignored."
        )
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["xAxis"]
    del common_options["yAxis"]
    del common_options["toolbox"]["feature"]["magicType"]
    series_options.update(
        {
            "type": "tree",
            "data": "#set_data#",
            "symbolSize": 7,
            "initialTreeDepth": 3,
            "layout": "orthogonal" if not radial else "radial",
            "orient": "horizontal" if not vertical else "vertical",
            "label": {
                "position": "left" if not vertical else "top",
                "rotate": 0 if not vertical else -90,
                "verticalAlign": "middle",
                "align": "right",
                "fontSize": 9,
            },
            "leaves": {
                "label": {
                    "position": "right" if not vertical else "bottom",
                    "rotate": 0 if not vertical else -90,
                    "verticalAlign": "middle",
                    "align": "left" if not vertical else "right",
                }
            },
            "emphasis": {"focus": "descendant"},
            "expandAndCollapse": True,
            "animationDuration": 550,
            "animationDurationUpdate": 750,
        }
    )
    common_options.update(
        {
            "tooltip": {"trigger": "item", "triggerOn": "mousemove"},
            "series": series_options,
        }
    )
    if radial:
        del common_options["series"]["orient"]
        del common_options["series"]["leaves"]
        del common_options["series"]["label"]

    if isinstance(data, dict):
        data = [data]

    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data, dump_whole=True),
        fields=["data"],
        options=common_options,
        order=order,
        title=title,
        option_modifications=option_modifications,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



stacked_area_data = [
    {
        "Email": 120,
        "Union Ads": 132,
        "Video Ads": 101,
        "Search Engine": 134,
        "Weekday": "Mon",
    },
    {
        "Email": 220,
        "Union Ads": 182,
        "Video Ads": 191,
        "Search Engine": 234,
        "Weekday": "Tue",
    },
    {
        "Email": 150,
        "Union Ads": 232,
        "Video Ads": 201,
        "Search Engine": 154,
        "Weekday": "Wed",
    },
    {
        "Email": 820,
        "Union Ads": 932,
        "Video Ads": 901,
        "Search Engine": 934,
        "Weekday": "Thu",
    },
    {
        "Email": 120,
        "Union Ads": 132,
        "Video Ads": 101,
        "Search Engine": 134,
        "Weekday": "Fri",
    },
    {
        "Email": 220,
        "Union Ads": 182,
        "Video Ads": 191,
        "Search Engine": 234,
        "Weekday": "Sat",
    },
    {
        "Email": 150,
        "Union Ads": 232,
        "Video Ads": 201,
        "Search Engine": 154,
        "Weekday": "Sun",
    },
]



bar_and_line_data = [
    {"day": "Mon", "Evaporation": 2.0, "Precipitation": 2.6, "Temperature": 2.0},
    {"day": "Tue", "Evaporation": 4.9, "Precipitation": 5.9, "Temperature": 2.2},
    {"day": "Wed", "Evaporation": 7.0, "Precipitation": 9.0, "Temperature": 3.3},
    {"day": "Thu", "Evaporation": 23.2, "Precipitation": 26.4, "Temperature": 4.5},
    {"day": "Fri", "Evaporation": 25.6, "Precipitation": 28.7, "Temperature": 6.3},
    {"day": "Sat", "Evaporation": 76.7, "Precipitation": 70.7, "Temperature": 10.2},
    {"day": "Sun", "Evaporation": 135.6, "Precipitation": 175.6, "Temperature": 20.3},
]



from typing import Dict, Optional, TYPE_CHECKING

import re
from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.app import App

import logging

logger = logging.getLogger(__name__)


class File(Resource):
    """File resource class"""

    _module_logger = logger
    resource_type = "file"
    plural = "files"
    alias_field = "name"
    elastic_supported = True

    def __init__(
        self,
        parent: "App",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[Dict] = None,
    ):
        params = dict(
            name=alias,
            fileName=re.sub("[^0-9a-zA-Z]+", "-", alias).lower() if alias else None,
            url="",
            contentType="",
            tags=[],
            metadata={},
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            params=params,
            check_params_before_creation=["name"],
            params_to_serialize=["metadata"],
        )

    async def delete(self):
        """Delete the report"""
        return await self._base_resource.delete()



from typing import Optional, TYPE_CHECKING
from shimoku.api.base_resource import Resource
from dataclasses import dataclass

if TYPE_CHECKING:
    from .business import Business

import logging

logger = logging.getLogger(__name__)


class BusinessInvitation(Resource):
    """BusinessInvitation resource class"""

    resource_type = "user/invitation"
    url_post_name = "user/invite"
    plural = "user/invitations"
    alias_field = "email"

    def __init__(
        self,
        parent: "Business",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            email="",
            roles=[],
            sendEmail=True,
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            check_params_before_creation=["email"],
            params=params,
        )

    async def delete(self) -> bool:
        return await self._base_resource.delete()


class BusinessUser(Resource):
    """BusinessUser resource class"""

    resource_type = "user"
    plural = "users"
    alias_field = "email"

    def __init__(
        self,
        parent: "Business",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = dict(
            id="",
            email="",
            name="",
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            # check_params_before_creation=["name"], Cant create this, Account is created by the system
            params=params,
        )

    async def delete(self) -> bool:
        return await self._base_resource.delete()



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def funnel_chart(
    self: "PlotLayer",
    order: int,
    names: str,
    values: str,
    data: Union[str, DataFrame, list[dict]],
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a scatter chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["xAxis"]
    del common_options["yAxis"]
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["legend"]["data"] = "#set_data#"
    common_options["tooltip"] = {"trigger": "item", "formatter": "{a} <br/>{b} : {c}%"}
    series_options.update(
        {
            "type": "funnel",
            "left": "10%",
            "top": "10%",
            "bottom": "10%",
            "width": "80%",
            "min": 0,
            "max": 100,
            "minSize": "0%",
            "maxSize": "100%",
            "sort": "descending",
            "gap": 2,
            "label": {"show": True, "position": "inside"},
            "labelLine": {"length": 10, "lineStyle": {"width": 1, "type": "solid"}},
            "itemStyle": {"borderColor": "#fff", "borderWidth": 1},
            "emphasis": {"label": {"fontSize": 20}},
        }
    )
    common_options.update({"series": series_options})

    await self._create_echart(
        fields=[names, {"name": names, "value": values}],
        options=common_options,
        data_mapping_to_tuples=await self._choose_data(order, data),
        option_modifications=option_modifications,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



import unittest
from os import getenv
from utils import initiate_shimoku
from shimoku.exceptions import APIError

s = initiate_shimoku()
business_id: str = getenv("BUSINESS_ID")
mock: bool = getenv("MOCK") == "TRUE"


class TestReport(unittest.TestCase):
    def test_get_report(self):
        s.set_workspace(uuid=business_id)
        s.set_menu_path("Report test path")

        s.plt.clear_menu_path()
        s.plt.html(html="<h1>test</h1>", order=0)

        report = s.menu_paths.get_menu_path_components(name="Report test path")[0]

        assert s.components.get_component(uuid=report["id"])

        s.components.delete_component(uuid=report["id"])

        if not mock:
            with self.assertRaises(APIError):
                s.components.get_component(uuid=report["id"])



from typing import Optional, Union, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.utils import validate_data_is_pandarable, ShimokuPalette
from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def scatter_chart(
    self: "PlotLayer",
    order: int,
    point_fields: Union[list[tuple[str, str]], tuple[str, str]],
    data: Union[str, DataFrame, list[dict]],
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    show_values: Union[list[str], str, None] = None,
    option_modifications: Optional[dict] = None,
    variant: Optional[str] = None,
):
    """Create a scatter chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    x_axis = common_options["xAxis"][0]
    del x_axis["data"]
    x_axis["type"] = "value"
    x_axis["scale"] = True
    common_options["yAxis"][0]["scale"] = True
    del common_options["toolbox"]["feature"]["magicType"]
    series_options["type"] = "scatter"
    series_options["symbolSize"] = 10
    del series_options["itemStyle"]
    del series_options["smooth"]
    await self._create_trend_chart(
        echart_options=common_options,
        series_options=series_options,
        option_modifications=option_modifications,
        axes=[],
        values=point_fields,
        data_mapping_to_tuples=await self._choose_data(order, data),
        order=order,
        x_axis_names=x_axis_name,
        y_axis_names=y_axis_name,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        show_values=show_values,
        variant=variant,
    )


async def scatter_with_effect_chart(
    self: "PlotLayer",
    order: int,
    x: str,
    y: str,
    data: Optional[Union[list[str], DataFrame, list[dict]]],
    x_axis_name: Optional[str] = None,
    y_axis_name: Optional[str] = None,
    title: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    padding: Optional[str] = None,
    effect_points: Optional[list] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a scatter chart"""
    data_mappings_to_tuples = await self._choose_data(order, data)
    if isinstance(data, str):
        data = self._shared_data[data]

    df: DataFrame = validate_data_is_pandarable(data)
    df = df[[x, y]].dropna()
    matrix = df.values

    common_options = get_common_echart_options()
    x_axis = common_options["xAxis"][0]
    del x_axis["data"]
    x_axis["type"] = "value"
    x_axis["scale"] = True
    common_options["yAxis"][0]["scale"] = True
    del common_options["toolbox"]["feature"]["magicType"]
    common_options["series"] = [
        {
            "data": [
                (point if isinstance(point, list) else list(matrix[point]))
                for point in effect_points
            ],
            "type": "effectScatter",
            "symbolSize": 20,
        },
        {
            "type": "scatter",
            "itemStyle": {
                "color": ShimokuPalette.CHART_C2.value,
            },
            "data": "#set_data#",
        },
    ]

    await self._create_echart(
        options=common_options,
        fields=[(x, y)],
        data_mapping_to_tuples=data_mappings_to_tuples,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        option_modifications=option_modifications,
    )



from shimoku import Client
from .data.table import table_data


def modal(shimoku_client: Client):
    prediction_header = (
        "<head>"
        "<style>"  # Styles title
        ".component-title{height:auto; width:100%; "
        "border-radius:16px; padding:16px;"
        "display:flex; align-items:center;"
        "background-color:var(--chart-C1); color:var(--color-white);}"
        "<style>.base-white{color:var(--color-white);}</style>"
        "</head>"  # Styles subtitle
        "<div class='component-title'>"
        "<div class='big-icon-banner'></div>"
        "<div class='text-block'>"
        "<h1>Predictions</h1>"
        "<p class='base-white'>"
        "Modal Test</p>"
        "</div>"
        "</div>"
    )
    shimoku_client.set_menu_path("Modal Test")

    shimoku_client.plt.set_modal(
        "Test modal", open_by_default=True, width=70, height=60
    )
    shimoku_client.plt.html(html=prediction_header, order=0)

    shimoku_client.plt.set_tabs_index(("Test", "Tab 1"), order=1)
    shimoku_client.plt.table(
        data=table_data,
        order=0,
        rows_size=3,
        title="Table test",
        categorical_columns=["filtA", "filtB"],
    )
    shimoku_client.plt.pop_out_of_tabs_group()

    shimoku_client.plt.html(html=prediction_header, order=2)
    shimoku_client.plt.pop_out_of_modal()

    shimoku_client.plt.html(html=prediction_header, order=0)
    shimoku_client.plt.set_tabs_index(("TestNoModal", "Table"), order=1)
    shimoku_client.plt.table(
        data=table_data,
        order=0,
        rows_size=3,
        title="Table test",
        categorical_columns=["filtA", "filtB"],
    )
    shimoku_client.plt.pop_out_of_tabs_group()

    shimoku_client.plt.modal_button(order=2, modal="Test modal", label="Open modal")



from shimoku import Client


def area(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "area-test")
    shimoku_client.plt.area(data="main data", x="date", order=0, variant="clean")
    shimoku_client.plt.area(
        data="main data", x="date", order=1, rows_size=2, cols_size=12
    )



from shimoku import Client
from .data.segmented_line import segmented_line_data


def segmented_line_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "Segmented Line Chart")
    marking_lines = [0, 50, 100, 150, 200, 300]
    range_colors = ["green", "yellow", "orange", "red", "purple", "maroon"]
    x_axis_name = "Date"
    y_axis_name = "AQI"

    shimoku_client.plt.segmented_line(
        data=segmented_line_data,
        order=0,
        x="date",
        y=["y", "y_displaced", "y_multiplied"],
        marking_lines=marking_lines,
    )

    shimoku_client.plt.segmented_line(
        data=segmented_line_data,
        order=1,
        x="date",
        y="y",
        title="Beijing's Air Quality Index",
        marking_lines=marking_lines,
        range_colors=range_colors,
        x_axis_name=x_axis_name,
        y_axis_name=y_axis_name,
        variant="clean",
    )



from typing import Optional
import logging

logger = logging.getLogger(__name__)


class FrontEndConnectionAPI:
    def __init__(self, js_snackbar: Optional[callable] = None):
        self.js_snackbar = js_snackbar

    def snackbar_error(self, message: str):
        """
        Show a snackbar with an error message.
        Only works in the browser.
        """
        self.js_snackbar(message, "error")

    def snackbar_info(self, message: str):
        """
        Show a snackbar with an info message.
        Only works in the browser.
        """
        self.js_snackbar(message, "info")

    def snackbar_success(self, message: str):
        """
        Show a snackbar with a success message.
        Only works in the browser.
        """
        self.js_snackbar(message, "success")


global_front_end_connection = FrontEndConnectionAPI()



import asyncio
import uuid
from typing import Optional, Callable, Dict, Union, Coroutine, Any, Type
import logging
from threading import Thread
from shimoku.utils import IN_BROWSER
from shimoku.api.client import ApiClient
import inspect

from copy import copy

logger = logging.getLogger(__name__)


def add_to_general_async_group(func: Callable):
    """
    This function adds the function to the general async group.
    """
    func.async_group = "general"
    return func


def add_to_individual_async_group(func: Callable):
    """
    This function creates an Async group for just one function,
    so that any number of the same function can be executed
    at the same time.
    """
    func.async_group = f"individual_{uuid.uuid4()}_{func.__name__}"
    return func


class AutoAsyncExecutionPool:
    """
    This class stores the arguments needed to execute
    the tasks in the task pool and the methods to execute them.
    It is used to avoid the use of async/await in the user's code.
    """

    ACTIONS_TEST = False

    def __init__(
        self,
        api_client: ApiClient,
    ):
        self.api_client = api_client
        self.ending_tasks: Dict[str, Coroutine] = {}
        self.task_pool = []
        self.free_context = {}
        # By default, set to true, to make the user aware that it is using the async configuration
        # (they will have to explicitly state it in their code)
        self.sequential = True
        self.universe = None
        self.in_async = False
        self._current_groups: list[Optional[str]] = []

    def clear(self):
        self.ending_tasks = {}
        self.task_pool = []
        self.free_context = {}
        self._current_groups = []

    async def execute_tasks(self) -> list:
        """
        This function executes the tasks in the task pool, and the ending tasks, if any.
        Lastly, it executes the last task in the task pool, and returns its result.
        """
        # IMPORTANT!! Nothing has to be dependent on this code as the sequential execution needs to keep working
        self.api_client.semaphore = asyncio.Semaphore(self.api_client.semaphore_limit)
        # To solve race conditions
        task_pool = copy(self.task_pool)
        self.task_pool.clear()
        results = await asyncio.gather(*task_pool)
        if self.ending_tasks:
            await asyncio.gather(*self.ending_tasks.values())
        self.clear()
        return list(results)

    def auto_async_func_call(
        self,
        func_self: Any,
        func: Callable,
        args: Optional[tuple] = None,
        kwargs: Optional[dict] = None,
        async_group: Optional[str] = None,
        name: Optional[str] = None,
    ) -> Union[Any, Coroutine]:
        """
        This function adds the function to the task pool, and executes it if the sequential execution is set to True.
        """
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        async def separate_final_execute(return_result: Optional[Any] = None):
            """
            This function executes the last task in the task pool, and returns its result.
            """
            if len(self.task_pool) > 0:
                await self.execute_tasks()
            self.task_pool.append(func(func_self, *args, **kwargs))
            task_result = (await self.execute_tasks())[0]
            if return_result is not None:
                return_result.append(task_result)
            return task_result

        needs_result = not async_group
        conflict = needs_result or (
            self._current_groups and async_group not in self._current_groups
        )

        if self.sequential or conflict:
            if IN_BROWSER or self.ACTIONS_TEST:
                # If in pyodide return the coroutine
                return separate_final_execute()
            # Try to retrieve the event loop, if it is not possible, create a new one
            try:
                # If the function is called from a coroutine, it will raise a RuntimeError
                # This is a workaround to avoid it
                asyncio.get_running_loop()
                return_result = []
                thread = Thread(
                    target=lambda result: asyncio.run(separate_final_execute(result)),
                    args=(return_result,),
                )
                thread.start()
                thread.join()
                return return_result[0] if return_result else None
            except RuntimeError:
                # If the function is called from the main synchronous thread, it will raise a RuntimeError
                return asyncio.run(separate_final_execute())

        # Copy the current context to make the execution independent and avoid the modification of the original context
        self.task_pool.append(func(copy(func_self), *args, **kwargs))
        self._current_groups.append(async_group)
        logger.info(f"{func.__name__ if not name else name} added to the task pool")

        if IN_BROWSER or self.ACTIONS_TEST:
            # If in pyodide return the coroutine
            return asyncio.sleep(0)


def decorate_class_to_auto_async(cls: type, async_pool: AutoAsyncExecutionPool) -> Type:
    """
    This function returns the class with all the methods decorated to be handled by the AutoAsyncExecutionPool.
    If the class has a method called _check_before_async_execution, it will be called before the async execution.
    If in PYODIDE execution all the methods will return a coroutine.
    """

    def decorate_to_auto_async(func: Callable):
        """
        This function decorates a method to be handled by the AutoAsyncExecutionPool.
        If the method has a return annotation, it will be executed and returned, if not it will be asummed that
        it can be executed with multiple tasks and will be added to the task pool.
        """

        def wrapper(self, *args, **kwargs):
            if inspect.iscoroutinefunction(func):
                if hasattr(self, "_check_before_async_execution"):
                    self._check_before_async_execution(
                        async_pool, func, *args, **kwargs
                    )
                return async_pool.auto_async_func_call(
                    func=func,
                    args=args,
                    kwargs=kwargs,
                    func_self=self,
                    async_group=func.async_group
                    if hasattr(func, "async_group")
                    else None,
                )
            result = func(self, *args, **kwargs)
            if IN_BROWSER or async_pool.ACTIONS_TEST:
                # If in pyodide return a wrapper coroutine
                async def result():
                    return result

                return result()
            return result

        return wrapper

    new_class = type(cls.__name__, (cls,), {})
    for attr_name in dir(cls):
        attr = getattr(cls, attr_name)
        if (
            callable(attr) and
            not attr_name.startswith("_") and
            not hasattr(attr, "__dataclass_fields__")
        ):
            setattr(new_class, attr_name, decorate_to_auto_async(attr))

    return new_class



import pandas as pd
from os import path
from shimoku import Client
from .data.default import data
from .data.pie import pie_data

data_path = path.join(path.dirname(path.abspath(__file__)), "data")


def doughnut(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "doughnut")

    shimoku_client.plt.doughnut(data=data, names="date", values="x", order=0)
    shimoku_client.plt.doughnut(data=pie_data, names="name", values="value", order=1)

    df = pd.read_csv(path.join(data_path, "test_stack_distribution.csv"))

    value_columns = [col for col in df.columns if col != "Segment"]
    df = df[["Segment"] + value_columns]

    aux_doughnut_data = pd.DataFrame(columns=["name", "value"])
    df_transposed = df.transpose().reset_index().drop(0)
    value_columns = [col for col in df_transposed.columns if col != "index"]
    aux_doughnut_data["value"] = df_transposed[value_columns].apply(
        lambda row: sum(row), axis=1
    )
    aux_doughnut_data["name"] = df_transposed["index"]
    shimoku_client.plt.doughnut(
        data=aux_doughnut_data,
        values="value",
        names="name",
        order=2,
        rows_size=3,
        cols_size=6,
    )



from os import getenv, path
from typing import Dict, List
from time import perf_counter
import unittest

import pandas as pd

from shimoku import Client
from utils import initiate_shimoku

from test_plot_api_functions.data.default import data
from test_plot_api_functions.data.zero_centered import zero_centered_data
from test_plot_api_functions.data.funnel import funnel_data
from test_plot_api_functions.data.heatmap import heatmap_data
from test_plot_api_functions.data.tree import tree_data
from test_plot_api_functions.data.noise import noise_data
from test_plot_api_functions.data.table import table_data
from test_plot_api_functions.data.horizontal_bar import horizontal_bar_data

from test_plot_api_functions.chart_and_indicators import chart_and_indicators
from test_plot_api_functions.infographics import infographics
from test_plot_api_functions.marked_line import marked_line_chart
from test_plot_api_functions.pie import pie
from test_plot_api_functions.radar import radar
from test_plot_api_functions.rose import rose
from test_plot_api_functions.scatter import scatter
from test_plot_api_functions.speed_gauge import speed_gauge
from test_plot_api_functions.stacked_area import stacked_area_chart
from test_plot_api_functions.stacked_bar import stacked_bar_chart
from test_plot_api_functions.summary_line import summary_line
from test_plot_api_functions.table import table
from test_plot_api_functions.tabs import tabs
from test_plot_api_functions.tree import tree
from test_plot_api_functions.treemap import treemap
from test_plot_api_functions.sankey import sankey
from test_plot_api_functions.heatmap import heatmap
from test_plot_api_functions.line import line
from test_plot_api_functions.area import area
from test_plot_api_functions.bar import bar
from test_plot_api_functions.horizontal_bar import horizontal_bar_chart
from test_plot_api_functions.zero_centered_bar import zero_centered_bar_chart
from test_plot_api_functions.funnel import funnel
from test_plot_api_functions.iframe import iframe
from test_plot_api_functions.bentobox import bentobox
from test_plot_api_functions.rainfall_area import rainfall_area
from test_plot_api_functions.rainfall_line import rainfall_line
from test_plot_api_functions.line_with_confidence_area import line_with_confidence_area
from test_plot_api_functions.scatter_with_effect import scatter_with_effect
from test_plot_api_functions.waterfall import waterfall
from test_plot_api_functions.bar_and_line_chart import bar_and_line_chart
from test_plot_api_functions.segmented_line_chart import segmented_line_chart
from test_plot_api_functions.segmented_area_chart import segmented_area_chart
from test_plot_api_functions.variants import variants
from test_plot_api_functions.annotation_chart import annotation_chart
from test_plot_api_functions.shimoku_gauges import shimoku_gauges
from test_plot_api_functions.gauge_indicator import gauge_indicators
from test_plot_api_functions.free_echarts import free_echarts
from test_plot_api_functions.input_form import input_form
from test_plot_api_functions.dynamic_conditional_and_auto_send_input_form import (
    dynamic_conditional_and_auto_send_input_form,
)
from test_plot_api_functions.get_input_forms import get_input_forms
from test_plot_api_functions.indicator import indicator
from test_plot_api_functions.indicator_color_by_value import indicator_color_by_value
from test_plot_api_functions.predictive_line import predictive_line
from test_plot_api_functions.html import html
from test_plot_api_functions.modal import modal
from test_plot_api_functions.doughnut import doughnut
from test_plot_api_functions.sunburst import sunburst
from test_plot_api_functions.table_with_lables import table_with_labels
from test_plot_api_functions.stacked_horizontal_bar import stacked_horizontal_bar_chart

data_path = path.join(
    path.dirname(path.abspath(__file__)), "test_plot_api_functions", "data"
)


class TestPlotApi(unittest.TestCase):
    def setUp(self):
        self.mock: bool = getenv("MOCK") == "TRUE"
        self.shimoku_client: Client = initiate_shimoku()
        self.shimoku_client.reuse_data_sets()
        self.workspace_id: str = getenv("BUSINESS_ID")
        self.shimoku_client.set_workspace(uuid=self.workspace_id)

    def test_delete_path(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        menu_path: str = "test-path"
        sub_path: str = "line-test"
        sub_path_2: str = "line-test-2"
        sub_path_3: str = "line-test-3"

        self.shimoku_client.set_menu_path(menu_path)
        self.shimoku_client.plt.clear_menu_path()

        self.shimoku_client.set_menu_path(menu_path, sub_path)

        self.shimoku_client.plt.line(data=data, x="date", order=0)
        self.shimoku_client.plt.line(data=data, x="date", order=1)

        reports: List[Dict] = self.shimoku_client.menu_paths.get_menu_path_components(
            name=menu_path
        )
        assert len(reports) == 2

        self.shimoku_client.plt.clear_menu_path()

        assert (
            len(self.shimoku_client.menu_paths.get_menu_path_components(name=menu_path))
            == 0
        )
        for i in range(10):
            self.shimoku_client.plt.line(data=data, x="date", order=i)

        self.shimoku_client.set_menu_path(menu_path, sub_path_2)
        self.shimoku_client.plt.line(data=data, x="date", order=0)

        self.shimoku_client.set_menu_path(menu_path, sub_path_3)
        self.shimoku_client.plt.line(data=data, x="date", order=0)

        assert (
            len(self.shimoku_client.menu_paths.get_menu_path_components(name=menu_path))
            == 12
        )

        self.shimoku_client.set_menu_path(menu_path, sub_path)
        self.shimoku_client.plt.clear_menu_path()
        assert (
            len(self.shimoku_client.menu_paths.get_menu_path_components(name=menu_path))
            == 2
        )

        self.shimoku_client.set_menu_path(menu_path)
        self.shimoku_client.plt.clear_menu_path()
        assert (
            len(self.shimoku_client.menu_paths.get_menu_path_components(name=menu_path))
            == 0
        )

        self.shimoku_client.set_menu_path("test")
        self.shimoku_client.menu_paths.delete_menu_path(name=menu_path)

        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_delete(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_menu_path("test-delete", "line-test")
        self.shimoku_client.plt.line(data=data, x="date", order=0)
        assert self.shimoku_client.menu_paths.get_menu_path_components(
            name="test-delete"
        )
        self.shimoku_client.plt.delete_chart_by_order(order=0)
        assert not self.shimoku_client.menu_paths.get_menu_path_components(
            name="test-delete"
        )
        self.shimoku_client.set_menu_path("test")
        self.shimoku_client.menu_paths.delete_menu_path(name="test-delete")
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_charts(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()

        self.shimoku_client.set_menu_path("test")
        self.shimoku_client.plt.set_shared_data(
            dfs={
                "main data": data,
                "zero centered data": zero_centered_data,
                "funnel data": funnel_data,
                "heatmap data": heatmap_data,
            },
            custom_data={
                "tree_data": tree_data,
            },
        )
        indicator(self.shimoku_client)
        indicator_color_by_value(self.shimoku_client)
        speed_gauge(self.shimoku_client)
        line(self.shimoku_client)
        predictive_line(self.shimoku_client)
        bar(self.shimoku_client)
        scatter(self.shimoku_client)
        area(self.shimoku_client)
        horizontal_bar_chart(self.shimoku_client)
        zero_centered_bar_chart(self.shimoku_client)
        funnel(self.shimoku_client)
        iframe(self.shimoku_client)
        bentobox(self.shimoku_client)
        radar(self.shimoku_client)
        pie(self.shimoku_client)
        rose(self.shimoku_client)
        doughnut(self.shimoku_client)
        html(self.shimoku_client)
        tree(self.shimoku_client)
        sunburst(self.shimoku_client)
        treemap(self.shimoku_client)
        sankey(self.shimoku_client)
        heatmap(self.shimoku_client)
        table(self.shimoku_client)
        self.shimoku_client.update_data_sets()
        table_with_labels(self.shimoku_client)
        self.shimoku_client.reuse_data_sets()
        annotation_chart(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_free_echarts(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_menu_path("test-free-echarts")
        stacked_data = pd.read_csv(path.join(data_path, "test_stack_distribution.csv"))
        self.shimoku_client.plt.set_shared_data(
            {
                "stacked data": stacked_data,
                "noise": noise_data,
                "table": table_data,
                "horizontal bar": horizontal_bar_data,
            }
        )
        stacked_bar_chart(self.shimoku_client)
        stacked_horizontal_bar_chart(self.shimoku_client)
        stacked_area_chart(self.shimoku_client)
        shimoku_gauges(self.shimoku_client)
        gauge_indicators(self.shimoku_client)
        free_echarts(self.shimoku_client)
        rainfall_area(self.shimoku_client)
        rainfall_line(self.shimoku_client)
        line_with_confidence_area(self.shimoku_client)
        scatter_with_effect(self.shimoku_client)
        waterfall(self.shimoku_client)
        bar_and_line_chart(self.shimoku_client)
        segmented_line_chart(self.shimoku_client)
        segmented_area_chart(self.shimoku_client)
        marked_line_chart(self.shimoku_client)
        variants(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_bento_box(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        infographics(self.shimoku_client)
        chart_and_indicators(self.shimoku_client)
        summary_line(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_filters(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_board("Test")
        self.shimoku_client.set_menu_path("test-filters")
        self.shimoku_client.plt.set_shared_data(
            dfs={
                "table": pd.DataFrame(table_data),
            }
        )
        self.shimoku_client.plt.filter(order=0, data="table", field="date")
        self.shimoku_client.plt.filter(order=1, data="table", field="x")
        self.shimoku_client.plt.filter(order=2, data="table", field="filtA")
        self.shimoku_client.plt.filter(
            order=3, data="table", field="filtB", multi_select=True, padding="0,0,0,8"
        )

        self.shimoku_client.plt.bar(
            data="table", order=4, cols_size=6, x="date", y=["x", "y"]
        )
        self.shimoku_client.plt.line(
            data="table", order=5, cols_size=6, x="date", y=["x", "y"]
        )
        self.shimoku_client.plt.stacked_bar(
            data="table", x="date", y=["x", "y"], order=6
        )
        self.shimoku_client.plt.table(
            data=table_data * 10,
            order=7,
            rows_size=3,
            title="Table test",
            categorical_columns=["filtA", "filtB"],
        )
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_tabs(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_board("Tabs dashboard")
        tabs(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_modals(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_board("Modals dashboard")
        modal(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_forms(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_board("Others")
        input_form(self.shimoku_client)
        dynamic_conditional_and_auto_send_input_form(self.shimoku_client)
        get_input_forms(self.shimoku_client)
        self.shimoku_client.run()
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_menu_order(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.workspaces.change_boards_order(
            uuid=self.workspace_id,
            boards=[
                "Default Name",
                "Test",
                "Modals dashboard",
                "Tabs dashboard",
                "Others",
                "---",
            ],
        )
        self.shimoku_client.workspaces.change_menu_order(
            uuid=self.workspace_id,
            menu_order=[
                "---",
                (
                    "test",
                    [
                        "line-test",
                        "bar-test",
                        "pie-test",
                        "area-test",
                        "scatter-test",
                        "radar-test",
                    ],
                ),
                (
                    "test-bentobox",
                    [
                        "chart-and-indicators",
                        "Infographics",
                    ],
                ),
                "test-free-echarts",
            ],
        )
        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_same_position_charts(self):
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()
        self.shimoku_client.set_menu_path("test-same-position", "no conflict path 1")

        self.shimoku_client.activate_async_execution()

        self.shimoku_client.plt.clear_menu_path()
        self.shimoku_client.plt.gauge_indicator(
            order=0,
            value=83,
            description="Síntomas coincidientes | Mareo, Dolor cervical",
            title="Sobrecarga muscular en cervicales y espalda",
        )

        self.shimoku_client.set_menu_path("test-same-position", "no conflict path 2")
        self.shimoku_client.plt.gauge_indicator(
            order=1,
            value=31,
            color=2,
            description="Síntomas coincidientes | Dolor cervical",
            title="Bruxismo",
        )

        self.shimoku_client.set_menu_path("test-same-position", "no conflict tabs")
        self.shimoku_client.plt.set_tabs_index(tabs_index=("tabs", "1"), order=0)
        self.shimoku_client.plt.gauge_indicator(
            order=0,
            value=83,
            description="Síntomas coincidientes | Mareo, Dolor cervical",
            title="Sobrecarga muscular en cervicales y espalda",
        )

        self.shimoku_client.plt.change_current_tab("2")
        self.shimoku_client.plt.gauge_indicator(
            order=1,
            value=31,
            color=2,
            description="Síntomas coincidientes | Dolor cervical",
            title="Bruxismo",
        )

        self.shimoku_client.plt.pop_out_of_tabs_group()
        self.shimoku_client.run()

        with self.assertRaises(RuntimeError):
            self.shimoku_client.set_menu_path("test-same-position", "conflict")
            self.shimoku_client.plt.gauge_indicator(
                order=0,
                value=83,
                description="Síntomas coincidientes | Mareo, Dolor cervical",
                title="Sobrecarga muscular en cervicales y espalda",
            )

            self.shimoku_client.plt.gauge_indicator(
                order=1,
                value=31,
                color=2,
                description="Síntomas coincidientes | Dolor cervical",
                title="Bruxismo",
            )
            self.shimoku_client.run()

        self.shimoku_client._async_pool.in_async = False

        with self.assertRaises(RuntimeError):
            self.shimoku_client.set_menu_path("test-same-position", "conflict")
            self.shimoku_client.plt.set_tabs_index(
                tabs_index=("conflict", "conflict"), order=0
            )
            self.shimoku_client.plt.gauge_indicator(
                order=0,
                value=83,
                description="Síntomas coincidientes | Mareo, Dolor cervical",
                title="Sobrecarga muscular en cervicales y espalda",
            )

            self.shimoku_client.plt.gauge_indicator(
                order=1,
                value=31,
                color=2,
                description="Síntomas coincidientes | Dolor cervical",
                title="Bruxismo",
            )
            self.shimoku_client.run()

        self.shimoku_client._async_pool.in_async = False

        self.shimoku_client.set_menu_path("test-same-position")
        self.shimoku_client.plt.clear_menu_path()

        assert 0 == len(
            self.shimoku_client.menu_paths.get_menu_path_components(
                name="test-same-position"
            )
        )

        self.shimoku_client.set_menu_path("test")
        self.shimoku_client.menu_paths.delete_menu_path(name="test-same-position")

        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_no_re_append_app_to_dashboard(self):
        if self.mock:
            return
        init_time = perf_counter()
        ini_calls = self.shimoku_client.get_api_calls_counter()

        self.shimoku_client.set_board("Test")
        self.shimoku_client.set_menu_path("test-no-reappend")
        self.shimoku_client.pop_out_of_menu_path()
        how_many_apps = len(
            self.shimoku_client.boards.get_board_menu_path_ids(name="Test")
        )
        self.shimoku_client.set_menu_path("test-no-reappend")
        self.shimoku_client.pop_out_of_menu_path()
        assert how_many_apps == len(
            self.shimoku_client.boards.get_board_menu_path_ids(name="Test")
        )

        s2 = initiate_shimoku()
        s2.set_workspace(self.workspace_id)
        s2.set_board("Test")
        s2.set_menu_path("test-no-reappend")
        s2.pop_out_of_menu_path()
        assert how_many_apps == len(s2.boards.get_board_menu_path_ids(name="Test"))

        self.shimoku_client.menu_paths.delete_menu_path(name="test-no-reappend")

        print(f"Total elapsed time: {perf_counter() - init_time:.2f} s")
        print(
            f"Number of api calls {self.shimoku_client.get_api_calls_counter() - ini_calls}"
        )

    def test_un_cached_modal(self):
        self.shimoku_client.disable_caching()
        self.shimoku_client.set_workspace(uuid=self.workspace_id)
        self.shimoku_client.set_menu_path("Un-cached modal test path")
        self.shimoku_client.plt.clear_menu_path()

        self.shimoku_client.plt.set_modal("Un-cached Modal")
        self.shimoku_client.plt.html(html="<h1>test</h1>", order=0)

        self.shimoku_client.run()

        res_html = self.shimoku_client.plt.get_component_by_order(0)
        assert res_html is not None
        assert res_html["chartData"][0]["value"] == "<h1>test</h1>"
        assert res_html["order"] == 0
        self.shimoku_client.plt.pop_out_of_modal()

        html_id = res_html["id"]
        modal_obj = self.shimoku_client.plt.get_modal("Un-cached Modal")
        assert html_id in modal_obj["properties"]["reportIds"]

        res_html = self.shimoku_client.plt.get_component_by_order(0)
        assert not res_html

        self.shimoku_client.enable_caching()


def main():
    # profiler = cProfile.Profile()
    # profiler.enable()
    my_test_plot_api = TestPlotApi()
    my_test_plot_api.setUp()
    for test in TestPlotApi.__dict__:
        if test.startswith("test_"):
            getattr(my_test_plot_api, test)()
    # profiler.disable()
    # stats = pstats.Stats(profiler)
    # stats.dump_stats('profile_output.prof')


if __name__ == "__main__":
    main()



from typing import Optional, TYPE_CHECKING

from pandas import DataFrame

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def sankey_chart(
    self: "PlotLayer",
    order: int,
    sources: str,
    targets: str,
    values: str,
    data: list[dict],
    option_modifications: Optional[dict] = None,
    title: Optional[str] = None,
    padding: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
):
    """Create a sankey chart"""
    df = DataFrame(data)
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["xAxis"]
    del common_options["yAxis"]
    del common_options["toolbox"]["feature"]["magicType"]
    del common_options["grid"]
    names = df[sources].to_list() + df[targets].to_list()
    series_options.update(
        {
            "type": "sankey",
            "layout": "none",
            "emphasis": {
                "focus": "adjacency",
            },
            "data": [
                {"name": name} for i, name in enumerate(names) if name not in names[:i]
            ],
            "links": "#set_data#",
        }
    )
    common_options["series"] = [series_options]
    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data),
        fields=[{"source": sources, "target": targets, "value": values}],
        options=common_options,
        order=order,
        title=title,
        option_modifications=option_modifications,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
    )



""""""
import json
import time

from os import getenv
from typing import Dict
from unittest import TestCase

import pandas as pd

from shimoku.exceptions import DataError
from shimoku import Client

from utils import initiate_shimoku

TEST_DATA_SET_NAME = "test-data-set"


class TestDataSets(TestCase):

    def sleep_cloud(self):
        if not self.shimoku_client.playground:
            time.sleep(5)

    def setUp(self):
        data: Dict = {
            "a": [x for x in range(10)],
            "b": [x ** 2 for x in range(10)],
        }
        self.data_oriented = sorted(
            pd.DataFrame(data).to_dict(orient="records"), key=lambda x: x["a"]
        )
        self.datajson: str = json.dumps(self.data_oriented)
        self.bad_dfs = [
            [{"a": 1, "b": 2}, {"a": 1, "b": 2, "c": 3}],
            [{"a": 1, "b": 2}, {"a": 1, "b": "c"}],
            [{"a": 1, "b": 2}, {"a": None, "b": 2}],
        ]
        self.shimoku_client: Client = initiate_shimoku()
        self.shimoku_client.set_workspace(getenv("BUSINESS_ID"))
        self.shimoku_client.set_menu_path(TEST_DATA_SET_NAME)
        self.shimoku_client.plt.clear_menu_path()
        
    def test_data_sets_CRUD(self):
        self.shimoku_client.data.append_to_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])
        for i in range(len(self.data_oriented)):
            assert data[i]["intField1"] == self.data_oriented[i]["a"]
            assert data[i]["intField2"] == self.data_oriented[i]["b"]
            assert data[i]["orderField1"] == i

        self.shimoku_client.data.append_to_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])
        for i in range(len(self.data_oriented)):
            assert data[i * 2]["intField1"] == self.data_oriented[i]["a"]
            assert data[i * 2]["intField2"] == self.data_oriented[i]["b"]
            assert data[i * 2]["orderField1"] == i

        self.shimoku_client.data.replace_data_from_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])
        assert len(data) == len(self.data_oriented)
        for i in range(len(self.data_oriented)):
            assert data[i]["intField1"] == self.data_oriented[i]["a"]
            assert data[i]["intField2"] == self.data_oriented[i]["b"]
            assert data[i]["orderField1"] == i

        self.shimoku_client.data.delete_data_set(name=TEST_DATA_SET_NAME)
        assert not self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)

    def test_delete_data_points(self):
        self.shimoku_client.data.append_to_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])
        for i in range(len(self.data_oriented)):
            assert data[i]["intField1"] == self.data_oriented[i]["a"]
            assert data[i]["intField2"] == self.data_oriented[i]["b"]
            assert data[i]["orderField1"] == i

        self.shimoku_client.data.delete_data_from_data_set(name=TEST_DATA_SET_NAME)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        assert not data

        self.shimoku_client.data.delete_data_set(name=TEST_DATA_SET_NAME)
        assert not self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)

    def test_linear_update(self):
        self.shimoku_client.data.append_to_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented[:5])
        self.sleep_cloud()

        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])

        for i in range(5):
            assert data[i]["intField1"] == self.data_oriented[i]["a"]
            assert data[i]["intField2"] == self.data_oriented[i]["b"]
            assert data[i]["orderField1"] == i

        modified_oriented_data = [{"a": dp["a"] + 10, "b": dp["b"] + 10} for dp in self.data_oriented]
        self.shimoku_client.data.linear_update_data_from_data_set(name=TEST_DATA_SET_NAME, data=modified_oriented_data)
        self.sleep_cloud()
        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        data = sorted(data, key=lambda x: x["orderField1"])

        for i in range(10):
            assert data[i]["intField1"] == modified_oriented_data[i]["a"]
            assert data[i]["intField2"] == modified_oriented_data[i]["b"]
            assert data[i]["orderField1"] == i

        modified_oriented_data = [{"a": 0, "b": 0}]
        self.shimoku_client.data.linear_update_data_from_data_set(name=TEST_DATA_SET_NAME, data=modified_oriented_data)
        self.sleep_cloud()

        data = self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
        assert len(data) == 1
        assert data[0]["intField1"] == 0
        assert data[0]["intField2"] == 0
        assert data[0]["orderField1"] == 0

        self.shimoku_client.data.delete_data_set(name=TEST_DATA_SET_NAME)
        assert not self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME)
         
    def test_bad_dfs(self):
        for i, bad_df in enumerate(self.bad_dfs):
            with self.assertRaises(DataError):
                self.shimoku_client.data.append_to_data_set(name=f"test-data-set_{i}", data=bad_df)

    def test_bad_append_to_existing_df(self):
        self.shimoku_client.data.replace_data_from_data_set(name=TEST_DATA_SET_NAME, data=self.data_oriented)
        self.sleep_cloud()
        print(self.shimoku_client.data.get_data_from_data_set(name=TEST_DATA_SET_NAME))
        for i, bad_df in enumerate(self.bad_dfs):
            bad_df_aux = [bad_df[1]]
            with self.assertRaises(DataError):
                self.shimoku_client.data.append_to_data_set(name=TEST_DATA_SET_NAME, data=bad_df_aux)



import pandas as pd
from shimoku import Client
from .data.bar_and_line import bar_and_line_data


def bar_and_line_chart(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "Bar and line chart")

    shimoku_client.plt.line_and_bar_charts(
        data=bar_and_line_data,
        order=0,
        x="day",
        bar_names=["Evaporation", "Precipitation"],
        line_names=["Temperature"],
        title="rainfall and temperature",
        x_axis_name="Day",
        line_axis_name="Temperature",
        line_suffix=" °C",
        bar_axis_name="Evaporation and precipitacion",
        bar_suffix=" ml",
        variant="minimal",
    )

    aux_data = pd.DataFrame(bar_and_line_data)

    aux_data["Temperature"] = aux_data["Temperature"] * 67
    aux_data["Evaporation"] = aux_data["Evaporation"] * 42
    aux_data["Precipitation"] = aux_data["Precipitation"] * 42

    shimoku_client.plt.line_and_bar_charts(
        data=aux_data,
        order=1,
        x="day",
        bar_names=["Evaporation", "Precipitation"],
        line_names=["Temperature"],
        title="rainfall and temperature",
        x_axis_name="Day",
        line_axis_name="Temperature",
        line_suffix=" °C",
        bar_axis_name="Evaporation and precipitacion",
        bar_suffix=" ml",
        variant="clean",
    )

    aux_data = pd.DataFrame(bar_and_line_data)

    aux_data["Temperature"] = aux_data["Temperature"] / 10000
    aux_data["Evaporation"] = aux_data["Evaporation"] * 542
    aux_data["Precipitation"] = aux_data["Precipitation"] * 542

    shimoku_client.plt.line_and_bar_charts(
        data=aux_data,
        order=2,
        x="day",
        bar_names=["Evaporation", "Precipitation"],
        line_names=["Temperature"],
        title="rainfall and temperature",
        x_axis_name="Day",
        line_axis_name="Temperature",
        line_suffix=" °C",
        bar_axis_name="Evaporation and precipitacion",
        bar_suffix=" ml",
    )



from shimoku import Client
from .data.default import data


def bar(shimoku_client: Client):
    shimoku_client.set_menu_path("test", "bar-test")
    shimoku_client.plt.bar(
        data=data,
        x="date",
        y=["x", "y"],
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )
    shimoku_client.plt.bar(data="main data", x="date", order=1)



from shimoku import Client
from .data.default import data
from .data.table import table_data
from .data.input_form import input_form_data


def tabs(shimoku_client: Client):
    shimoku_client.set_menu_path("test-tabs")
    # s.plt.clear_menu_path()
    shimoku_client.plt.set_shared_data(dfs={"main data": data})

    def _test_bentobox():
        shimoku_client.plt.set_bentobox(cols_size=8, rows_size=3)
        data_indic = [
            {
                "description": "",
                "title": "Estado",
                "value": "Abierto",
            },
        ]
        shimoku_client.plt.indicator(
            data=data_indic,
            order=0,
            rows_size=10,
            cols_size=12,
        )

        shimoku_client.plt.indicator(
            data=data_indic,
            order=1,
            rows_size=10,
            cols_size=12,
        )

        shimoku_client.plt.bar(
            data="main data",
            x="date",
            order=2,
            rows_size=26,
            cols_size=24,
        )
        shimoku_client.plt.pop_out_of_bentobox()

    shimoku_client.plt.set_tabs_index(tabs_index=("Deepness 0", "Bento box"), order=0)
    _test_bentobox()

    shimoku_client.plt.change_current_tab("Table")
    shimoku_client.plt.table(
        title="Test-table",
        data=table_data,
        order=0,
        categorical_columns=["filtA", "filtB"],
        initial_sort_column="date",
        search=True,
    )

    shimoku_client.plt.change_current_tab("line test")
    shimoku_client.plt.line(data="main data", x="date", order=0)

    shimoku_client.plt.change_current_tab("Bar 1")
    shimoku_client.plt.bar(
        data="main data",
        x="date",
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )

    shimoku_client.plt.change_current_tab("Input Form")
    shimoku_client.plt.input_form(order=0, options=input_form_data)

    indicators_data = {
        "description": "",
        "title": "",
        "value": "INDICATOR CHANGED!",
        "color": "warning",
    }
    indicators_data_2 = {
        "description": "",
        "title": "",
        "value": "INDICATOR CHANGED!",
        "color": "main",
    }
    shimoku_client.plt.change_current_tab("Indicators 2")
    shimoku_client.plt.indicator(data=indicators_data_2, order=0)
    shimoku_client.plt.change_current_tab("Indicators 1")
    shimoku_client.plt.indicator(data=indicators_data, order=0)

    shimoku_client.plt.set_tabs_index(
        tabs_index=("Deepness 1", "Bento box"),
        order=1,
        parent_tabs_index=("Deepness 0", "Indicators 1"),
        sticky=False,
        just_labels=True,
    )
    _test_bentobox()
    shimoku_client.plt.change_current_tab("Indicators 1")

    for i in range(2, 5):
        shimoku_client.plt.set_tabs_index(
            (f"Deepness {i}", "Indicators 2"),
            order=1,
            parent_tabs_index=(f"Deepness {i - 1}", "Indicators 1"),
            sticky=False,
            just_labels=True,
        )
        shimoku_client.plt.indicator(data=indicators_data_2, order=0)
        shimoku_client.plt.change_current_tab("Indicators 1")
        shimoku_client.plt.indicator(data=indicators_data, order=0)

    shimoku_client.plt.set_tabs_index(
        ("Bar deep 1", "Bar 1"),
        order=1,
        parent_tabs_index=("Deepness 0", "Bar 1"),
        sticky=False,
        just_labels=True,
    )
    shimoku_client.plt.bar(
        data="main data",
        x="date",
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )
    shimoku_client.plt.set_tabs_index(
        ("Bar deep 2", "Bar 2"),
        order=1,
        parent_tabs_index=("Bar deep 1", "Bar 1"),
        sticky=False,
        just_labels=True,
    )
    shimoku_client.plt.bar(
        data="main data",
        x="date",
        y="y",
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=2,
        rows_size=2,
        cols_size=12,
    )
    shimoku_client.plt.change_current_tab("Line 1")
    shimoku_client.plt.line(
        data="main data",
        x="date",
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )
    shimoku_client.plt.change_current_tab("Line 2")
    shimoku_client.plt.line(
        data="main data",
        x="date",
        x_axis_name="Date",
        y_axis_name="Revenue",
        order=0,
        rows_size=2,
        cols_size=12,
    )
    shimoku_client.plt.pop_out_of_tabs_group()



from typing import Optional, TYPE_CHECKING, TypedDict

import asyncio

from shimoku.api.base_resource import Resource

if TYPE_CHECKING:
    from shimoku.api.resources.universe import Universe

import logging

logger = logging.getLogger(__name__)

MAX_CODE_FRAGMENT_SIZE = 8000


class ActionScript(Resource):
    _module_logger = logger
    resource_type = "script"
    plural = "scripts"

    class ActionScriptParams(TypedDict):
        order: int
        codeFragment: str

    def __init__(
        self,
        parent: "Action",
        uuid: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = ActionScript.ActionScriptParams(
            order=0,
            codeFragment="",
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            check_params_before_creation=["order"],
            params=params,
        )

    async def delete(self):
        """Deletes the action script."""
        return await self._base_resource.delete()


class Action(Resource):
    _module_logger = logger
    resource_type = "action"
    alias_field = "name"
    plural = "actions"

    class ActionParams(TypedDict):
        name: Optional[str]
        description: Optional[str]
        actionTemplateId: Optional[str]
        universeApiKeyId: Optional[str]
        pythonLibraries: Optional[list[str]]

    def __init__(
        self,
        parent: "Universe",
        uuid: Optional[str] = None,
        alias: Optional[str] = None,
        db_resource: Optional[dict] = None,
    ):
        params = Action.ActionParams(
            name=alias,
            description=None,
            actionTemplateId=None,
            universeApiKeyId=None,
            pythonLibraries=None,
        )

        super().__init__(
            parent=parent,
            uuid=uuid,
            db_resource=db_resource,
            children=[ActionScript],
            check_params_before_creation=["name"],
            params=params,
        )

    async def delete(self):
        """Deletes the action."""
        return await self._base_resource.delete()

    async def update(self):
        """Updates the action."""
        return await self._base_resource.update()

    async def get_code(self) -> str:
        """Gets the code of the action."""
        scripts = sorted(
            await self._base_resource.get_children(ActionScript),
            key=lambda s: s["order"],
        )
        return "".join([script["codeFragment"] for script in scripts])

    async def upload_code(self, code: str):
        """Uploads the code of the action."""
        current_action_scripts = await self._base_resource.get_children(ActionScript)
        if current_action_scripts:
            logger.warning(
                "The action already contains code. The current code will be deleted."
            )
            await asyncio.gather(
                *[script.delete() for script in current_action_scripts]
            )
        await asyncio.gather(
            *[
                self._base_resource.create_child(
                    ActionScript,
                    order=i,
                    codeFragment=code_fragment,
                )
                for i, code_fragment in enumerate(
                    code[i : i + MAX_CODE_FRAGMENT_SIZE]
                    for i in range(0, len(code), MAX_CODE_FRAGMENT_SIZE)
                )
            ]
        )

    async def delete_code(self):
        """Deletes the code of the action."""
        current_action_scripts = await self._base_resource.get_children(ActionScript)
        await asyncio.gather(*[script.delete() for script in current_action_scripts])



""""""
from os import getenv
import inspect

from shimoku import Client

api_key: str = getenv("API_TOKEN")
universe_id: str = getenv("UNIVERSE_ID")
business_id: str = getenv("BUSINESS_ID")
environment: str = getenv("ENVIRONMENT")
verbosity: str = getenv("VERBOSITY")


s = Client(
    access_token=api_key,
    universe_id=universe_id,
    environment=environment,
    verbosity=verbosity,
)
s.set_workspace(uuid=business_id)

s.activate_async_execution()

components = [f for _, f in s.html_components.__dict__.items() if callable(f)]

parameters = dict(
    background_url="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    background="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    image="https://images.unsplash.com/photo-1553356084-58ef4a67b2a7?ixlib=rb-4.0.3&ixid=MnwxMjA3fDB8MHxwaG90by1wYWdlfHx8fGVufDB8fHx8&auto=format&fit=crop&w=774&q=80",
    image_option="animated",
    title="Lorem ipsum",
    subtitle="Lorem ipsum dolor sit amet",
    href="https://shimoku.com",
    text="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore "
    "et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
    "aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum "
    "dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia "
    "deserunt mollit anim id est laborum.",
    main_text="Lorem ipsum",
    internal_text="Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore "
    "et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut "
    "aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum "
    "dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia "
    "deserunt mollit anim id est laborum.",
    note="Lorem ipsum",
    button_panel="Lorem ipsum",
    button_text="Lorem ipsum",
    symbol_name="insights",
    h1="Lorem ipsum",
    line="Lorem ipsum",
    modal_title="Lorem ipsum",
    background_color="var(--color-primary)",
)

for component_function in components:
    function_signature = inspect.signature(component_function)
    call_parameters = {
        parameter[0]: parameter[1]
        for parameter in parameters.items()
        if parameter[0] in function_signature.parameters
    }
    print(component_function)
    print(function_signature.parameters)
    # If the parameters dont exist in the function signature, skip it
    if not call_parameters:
        continue
    print(f"passed {component_function.__name__}")
    html = component_function(**call_parameters)
    s.set_menu_path("Components test", f"{component_function.__name__}")
    s.plt.html(html=html, order=0)

s.run()



from shimoku import Client


def variants(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "Variants")

    shimoku_client.plt.stacked_bar(
        data="stacked data",
        order=0,
        cols_size=6,
        x="Segment",
        variant="clean shadow",
        show_values="all",
    )
    shimoku_client.plt.stacked_horizontal_bar(
        data="stacked data",
        order=1,
        cols_size=6,
        x="Segment",
        variant="minimal shadow",
        show_values="all",
    )
    shimoku_client.plt.bar(
        data="table",
        order=2,
        cols_size=6,
        x="date",
        variant="shadow",
        show_values=["x"],
    )
    shimoku_client.plt.horizontal_bar(
        data="horizontal bar",
        x="name",
        order=3,
        cols_size=6,
        show_values=["value"],
        variant="clean shadow",
    )
    shimoku_client.plt.bar(
        data="table",
        order=4,
        cols_size=6,
        x="date",
        variant="clean thin",
        show_values=["x"],
    )
    shimoku_client.plt.horizontal_bar(
        data="horizontal bar",
        x="name",
        order=5,
        cols_size=6,
        show_values=["value"],
        variant="clean thin",
    )
    shimoku_client.plt.stacked_bar(
        data="stacked data",
        order=6,
        cols_size=6,
        x="Segment",
        variant="thin",
        show_values="all",
    )
    shimoku_client.plt.stacked_horizontal_bar(
        data="stacked data",
        order=7,
        cols_size=6,
        x="Segment",
        variant="clean thin",
        show_values="all",
    )
    shimoku_client.plt.area(
        data="table",
        order=8,
        cols_size=3,
        rows_size=2,
        x="date",
        y="x",
        variant="minimal",
    )
    shimoku_client.plt.line(
        data="table",
        order=9,
        cols_size=3,
        rows_size=2,
        x="date",
        y="x",
        variant="minimal",
    )



import asyncio
from typing import Optional, Union

from copy import copy

from shimoku.api.resources.app import App
from shimoku.api.resources.universe import Universe
from shimoku.api.resources.activity import Activity
from shimoku.api.resources.file import File
from shimoku.api.resources.activity_template import ActivityTemplate

from shimoku.exceptions import AIFunctionError, ShimokuFileError, ActivityError

import logging
from shimoku.execution_logger import log_error, ClassWithLogging

logger = logging.getLogger(__name__)


def check_app_is_set(self: Union["AILayer", "AIFunctionMethods"]):
    """Check that the app is set"""
    if self._app is None:
        log_error(
            logger,
            "Menu path not set. Please use set_menu_path() method first.",
            AttributeError,
        )


def get_model_metadata(model: File) -> dict:
    """Get the metadata of a model
    :param model: Model
    :return: The metadata of the model
    """
    metadata: dict = copy(model["metadata"])
    for tag in model["tags"]:
        if tag.startswith("creator_ai_function_version:"):
            metadata["creator_ai_function_version"] = tag[
                len("creator_ai_function_version:") :
            ]
        elif tag.startswith("creator_ai_function_id"):
            metadata["creator_ai_function_id"] = tag[len("creator_ai_function_id:") :]
        elif tag.startswith("creator_ai_function"):
            metadata["creator_ai_function"] = tag[len("creator_ai_function:") :]
    return metadata


def get_output_file_metadata(file: File):
    """Get the metadata of an output file
    :param file: File
    :return: The metadata of the output file
    """
    metadata: dict = copy(file["metadata"])
    for tag in file["tags"]:
        if tag.startswith("creator_ai_function_version:"):
            metadata["creator_ai_function_version"] = tag[
                len("creator_ai_function_version:") :
            ]
        elif tag.startswith("creator_ai_function:"):
            metadata["creator_ai_function"] = tag[len("creator_ai_function:") :]
        elif tag.startswith("model_name:"):
            metadata["model_name"] = tag[len("model_name:") :]
    return metadata


def get_output_file_name(
    activity_template: ActivityTemplate, file_name: str, run_id: str
) -> str:
    """Get the name of an output file of a ai_function
    :param activity_template: Activity template of the ai_function
    :param file_name: Name of the file
    :param run_id: Id of the run
    :return: The name of the output file
    """
    name = activity_template["name"]
    version = activity_template["version"]
    return f"shimoku_generated_output_file_{name}_{version}_{run_id}_{file_name}"


def get_input_file_name(file_name: str):
    """Get the name of an input file of a ai_function
    :param file_name: Name of the file
    :return: The name of the input file
    """
    return f"shimoku_generated_input_file_{file_name}"


async def check_and_get_model(app: App, model_name: str) -> File:
    """Check that a model exists and get it
    :param app: App
    :param model_name: Name of the model
    """
    app_files = await app.get_files()
    for file in app_files:
        if "shimoku_generated" not in file["tags"] or "ai_model" not in file["tags"]:
            continue
        metadata = get_model_metadata(file)
        if (
            file["name"] == "shimoku_generated_model_" + model_name
            and metadata["model_name"] == model_name
        ):
            return file
    log_error(logger, f"The model ({model_name}) does not exist", ShimokuFileError)


class AIFunctionMethods(ClassWithLogging):
    _module_logger = logger
    _use_info_logging = True

    def __init__(self, app: App, activity_template: ActivityTemplate, run_id: str):
        self._app: App = app
        self._private_ai_function: ActivityTemplate = activity_template
        self._run_id: str = run_id

    async def _create_output_file(
        self,
        activity_template: ActivityTemplate,
        file_name: str,
        file: Union[bytes, tuple[bytes, dict]],
        run_id: str,
        model_name: Optional[str] = None,
    ) -> str:
        """Create an output file of a ai_function
        :param activity_template: Activity template of the ai_function
        :param file: File to be uploaded
        :param run_id: Id of the run
        :return: The uuid of the created file
        """
        metadata = {}
        if isinstance(file, tuple):
            file, metadata = file

        complete_file_name = get_output_file_name(activity_template, file_name, run_id)

        tags = [
            "shimoku_generated",
            "ai_output_file",
            f'creator_ai_function:{activity_template["name"]}',
            f'creator_ai_function_version:{activity_template["version"]}',
        ]
        if model_name is not None:
            tags.append(f"model_name:{model_name}")

        file = await self._app.create_file(
            name=complete_file_name,
            file_object=file,
            tags=tags,
            metadata={"file_name": file_name, "run_id": run_id, **metadata},
        )
        logger.info(f"Created output file ({file_name})")
        return file["id"]

    async def create_model(
        self, model_name: str, model: bytes, metadata: Optional[dict] = None
    ):
        """Create a model
        :param model_name: Name of the model
        :param model: Model to be uploaded
        :param metadata: Metadata of the model
        """
        if not model_name or not isinstance(model_name, str):
            log_error(logger, "Model name has to be a non-empty string", ValueError)

        file_name = f"shimoku_generated_model_{model_name}"

        check_app_is_set(self)
        await self._app.create_file(
            name=file_name,
            file_object=model,
            tags=[
                "shimoku_generated",
                "ai_model",
                f"creator_ai_function:{self._private_ai_function['name']}",
                f"creator_ai_function_version:{self._private_ai_function['version']}",
                f"creator_ai_function_id:{self._private_ai_function['id']}",
            ],
            metadata={
                "model_name": model_name,
                "run_id": self._run_id,
                **(metadata or {}),
            },
        )
        logger.info(f"Created model ({model_name})")

    async def get_model(self, model_name: str) -> tuple[bytes, dict]:
        """Get a model
        :param model_name: Name of the model
        :return: The model if it exists
        """
        check_app_is_set(self)
        model_file: File = await check_and_get_model(self._app, model_name)
        return await self._app.get_file_object(model_file["id"]), get_model_metadata(
            model_file
        )

    async def create_output_files(
        self,
        files: dict[str, Union[bytes, tuple[bytes, dict]]],
        model_name: Optional[str] = None,
    ):
        """Create output files of a ai_function
        :param files: Files to be uploaded
        :param model_name: Name of the model used
        """
        check_app_is_set(self)
        if model_name is not None:
            await check_and_get_model(self._app, model_name)
        else:
            model_name = ""

        await asyncio.gather(
            *[
                self._create_output_file(
                    self._private_ai_function, file_name, file, self._run_id, model_name
                )
                for file_name, file in files.items()
            ]
        )


async def get_universe_api_key(
    universe: Universe, universe_api_key: Optional[str]
) -> str:
    """Get the universe API key or create one if none exists
    :param universe: Universe
    :param universe_api_key: Universe API key
    :return: Universe API key
    """
    if universe_api_key is None:
        universe_api_keys = await universe.get_universe_api_keys()
        if len(universe_api_keys) == 0:
            return (await universe.create_universe_api_key("AI functions key"))["id"]
        return universe_api_keys[0]["id"]
    return universe_api_key


async def get_activity_template(
    universe: Universe, name: str, version: Optional[str]
) -> ActivityTemplate:
    """Get an activity template
    :param universe: Universe
    :param name: Name of the activity template
    :param version: Version of the activity template
    :return: Activity template
    """
    if version is not None:
        activity_template = await universe.get_activity_template(
            name_version=(name, version)
        )
    else:
        activity_templates = await universe.get_activity_templates()
        activity_templates = [
            activity_template
            for activity_template in activity_templates
            if activity_template["name"] == name
        ]
        activity_template = (
            activity_templates[-1] if len(activity_templates) > 0 else None
        )
    if activity_template is None:
        log_error(
            logger,
            f"The AI function ({name} {'v' + (version if version is not None else '')}) does not exist",
            AIFunctionError,
        )
    return activity_template


async def get_activity_from_template(
    app: App,
    activity_template: ActivityTemplate,
    universe_api_key: Optional[str] = None,
    create_if_not_exists: bool = True,
) -> Activity:
    """Get an activity from an activity template
    :param app: App
    :param activity_template: Activity template
    :param universe_api_key: Universe API key
    :param create_if_not_exists: Create the activity if it does not exist
    :return: Activity
    """
    universe_api_key = await get_universe_api_key(
        activity_template.parent, universe_api_key
    )

    name = activity_template["name"]
    activity_name = "shimoku_generated_activity_" + name

    activity: Optional[Activity] = await app.get_activity(name=activity_name)
    if activity is None:
        if not create_if_not_exists:
            log_error(
                logger,
                f"The activity for the AI function ({name}) does not exist",
                AIFunctionError,
            )
        activity: Activity = await app.create_activity(
            name=activity_name,
            template_id=activity_template["id"],
            universe_api_key=universe_api_key,
        )
        logger.info(f"Created activity for ai_function {name}")

    return activity


async def fill_input_for_run_metadata(
    key: str,
    value: any,
    args: dict,
    files: dict,
    input_settings: dict,
    app_files: list[File],
):
    """Fill the files and args of a run metadata
    :param key: Key of the input
    :param value: Value of the input
    :param args: Args of the run metadata
    :param files: Files of the run metadata
    :param input_settings: Settings of the input
    :param app_files: Files of the app
    """
    if key not in input_settings:
        return
    if input_settings[key]["datatype"] == "file":
        filtered_files: list[File] = [file for file in app_files if file["id"] == value]
        if len(filtered_files) > 1:
            log_error(logger, f"More than one file with ID {value} found", ValueError)
        elif len(filtered_files) == 0:
            value = "# file not found #"
        else:
            file = filtered_files[0]
            value = file["metadata"]
        files[key] = value
    else:
        args[key] = value


async def fill_run_metadata(
    app: App,
    run_metadata: dict,
    run_id: str,
    ai_function_name: str,
    ai_function_version: str,
):
    """Fill the metadata of a run
    :param app: App
    :param run_metadata: Metadata of the run to fill
    :param run_id: ID of the run to fill
    :param ai_function_name: Name of the ai_function used
    :param ai_function_version: Version of the ai_function used
    """
    activity_template: ActivityTemplate = await get_activity_template(
        app.parent.parent, ai_function_name, ai_function_version
    )
    activity: Activity = await get_activity_from_template(
        app, activity_template, create_if_not_exists=False
    )
    activity_runs: list[Activity.Run] = await activity.get_runs()
    runs = [run for run in activity_runs if run["id"] == run_id]
    if len(runs) > 1:
        log_error(logger, f"More than one run with ID {run_id} found", ValueError)
    if len(runs) == 0:
        run_metadata["input"] = "# Run not found #"
        return
    run: Activity.Run = runs[0]
    input_settings = activity_template["inputSettings"]
    args = {}
    files = {}
    await asyncio.gather(
        *[
            fill_input_for_run_metadata(
                k, v, args, files, input_settings, await app.get_files()
            )
            for k, v in run["settings"].items()
        ]
    )
    run_metadata["input"]["args"] = args
    run_metadata["input"]["files"] = files


async def check_and_store_output_file(
    file: File,
    files_by_run_id: list[dict],
    run_id: Optional[str] = None,
    model_metadata: Optional[dict] = None,
    activity_template: Optional[ActivityTemplate] = None,
    file_name: Optional[str] = None,
    get_objects: bool = False,
):
    """Check if a file is an output file then add it to the list of output files
    :param model_metadata: Metadata of the model used
    :param activity_template: ai_function used
    :param file: File to check
    :param files_by_run_id: dictionary of output files by run ID
    :param run_id: ID of the run to check
    :param file_name: Name of the file to check
    :param get_objects: Get the objects of the files instead of their metadata
    """
    if "shimoku_generated" not in file["tags"] or "ai_output_file" not in file["tags"]:
        return

    output_file_metadata: dict = {}
    aux_output_file_metadata: dict = get_output_file_metadata(file)
    output_file_metadata.update(aux_output_file_metadata)

    output_file_name: str = output_file_metadata.pop("file_name")
    output_file_model_name: Optional[str] = output_file_metadata.get("model_name")
    ai_function_name: str = output_file_metadata.pop("creator_ai_function")
    ai_function_version: str = output_file_metadata.pop("creator_ai_function_version")

    if (
        (
            activity_template
            and (
                ai_function_name != activity_template["name"]
                or ai_function_version != activity_template["version"]
            )
        )
        or (model_metadata and output_file_model_name != model_metadata["model_name"])
        or (file_name and output_file_name != file_name)
    ):
        return

    output_file_run_id: str = output_file_metadata.pop("run_id")
    if run_id and output_file_run_id != run_id:
        return

    run_metadata = {"run_id": output_file_run_id}
    if output_file_run_id not in [run["run_id"] for run in files_by_run_id]:
        if not model_metadata and output_file_model_name is not None:
            run_metadata.update({"model_name": output_file_model_name})
        if not activity_template:
            run_metadata.update(
                {
                    "ai_function_name": ai_function_name,
                    "ai_function_version": ai_function_version,
                }
            )
        run_metadata.update({"input": {"args": [], "files": []}, "output_files": {}})
        files_by_run_id.append(run_metadata)
    else:
        run_metadata = [
            run_d for run_d in files_by_run_id if run_d["run_id"] == output_file_run_id
        ][0]

    if get_objects:
        output_file_metadata["object"] = await file.parent.get_file_object(file["id"])

    output_file_metadata = {
        k: v
        for k, v in output_file_metadata.items()
        if k not in ["creator_ai_function", "creator_ai_function_version"]
    }
    run_metadata["output_files"][output_file_name] = output_file_metadata

    await fill_run_metadata(
        file.parent,
        run_metadata,
        output_file_run_id,
        ai_function_name,
        ai_function_version,
    )


async def get_runs_message(
    runs: list[Activity.Run], wf_input_settings: Union[dict, list]
) -> list[str]:
    """Get the output from a list of runs
    :param runs: list of runs
    :param wf_input_settings: Input settings of the ai_function/s
    :return: Message of the run
    """
    message = []
    # No need to use 'gather' here because logs are already cached from ordering the runs
    for i, run in enumerate(runs[::-1]):
        if isinstance(wf_input_settings, list):
            input_settings = wf_input_settings[i]
        else:
            input_settings = wf_input_settings
        run_settings = {
            k: v
            for k, v in run["settings"].items()
            if k in input_settings and input_settings[k]["datatype"] != "file"
        }
        message.extend(
            [
                "",
                f' - Run {run["id"]}:',
                f'   Settings: {", ".join([f"{key}: {value}" for key, value in run_settings.items()])}',
                "   Logs:",
                *[
                    f'     - {log["severity"]} | {log["message"]}, at {log["dateTime"]}'
                    for log in (await run.get_logs())
                ],
            ]
        )

    message.extend(["", "///////////////////////////////////////////////////", ""])
    return message


class AILayer(ClassWithLogging):
    _module_logger = logger
    _use_info_logging = True

    def __init__(
        self,
        access_token: str,
        universe: Universe,
        app: App,
    ):
        self._app: App = app
        self._universe: Universe = universe
        self._access_token: str = access_token
        self._private_ai_function_and_run_id: Optional[
            tuple[ActivityTemplate, str]
        ] = None
        if app is not None and universe["id"] != app.parent.parent["id"]:
            log_error(
                logger,
                f"Menu path ({str(app)}) does not belong to the universe ({str(universe)})",
                AIFunctionError,
            )

    async def _check_run_id_exists(
        self, activity_template: ActivityTemplate, run_id: str
    ):
        """Check if a run ID exists for the activity of a ai_function
        :param activity_template: Activity template to check
        :param run_id: ID of the run to check
        """
        activity: Activity = await get_activity_from_template(
            self._app, activity_template, create_if_not_exists=False
        )
        if not await activity.get_run(run_id):
            log_error(
                logger,
                f"Run ({run_id}) does not exist for the activity ({activity_template['name']})",
                ActivityError,
            )

    async def check_for_private_access(self, ai_function_id: str, run_id: str):
        """Check if the credentials allow for private ai_function methods"""
        universe_api_keys = [
            uak["id"] for uak in await self._universe.get_universe_api_keys()
        ]
        if self._access_token not in universe_api_keys:
            log_error(
                logger,
                "The access key must be a universe API key to use private AI function methods",
                AIFunctionError,
            )

        activity_template: ActivityTemplate = (
            await self._universe.get_activity_template(ai_function_id)
        )

        check_app_is_set(self)
        await self._check_run_id_exists(activity_template, run_id)

        self._private_ai_function_and_run_id = (activity_template, run_id)

    def get_private_ai_function_methods(self) -> AIFunctionMethods:
        """Get the private ai_function methods if the credentials allow for it"""
        if not self._private_ai_function_and_run_id:
            log_error(
                logger,
                "Credentials must be set to use private AI function methods, "
                "please call check_for_private_access first",
                AIFunctionError,
            )
        activity_template, run_id = self._private_ai_function_and_run_id
        return AIFunctionMethods(self._app, activity_template, run_id)

    async def delete_model(self, model_name: str):
        """Delete a model
        :param model_name: Name of the model
        """
        check_app_is_set(self)
        model_file: File = await check_and_get_model(self._app, model_name)
        await self._app.delete_file(model_file["id"])
        logger.info(f"Deleted model ({model_name})")

    async def get_last_executions_with_logs(
        self, ai_function: str, version: Optional[str] = None, how_many: int = 1
    ):
        """Get the logs of the executions of a ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param how_many: Number of executions to get
        :return: The logs of the ai_function
        """
        activity_template: ActivityTemplate = await get_activity_template(
            self._universe, ai_function, version
        )
        activity: Activity = await get_activity_from_template(
            self._app, activity_template, create_if_not_exists=False
        )
        runs: list[Activity.Run] = await activity.get_runs(how_many)
        return [run.cascade_to_dict() for run in runs]

    async def get_output_files_by_model(
        self,
        model_name: str,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """Get output files for the executions of ai_functions with a given model
        :param model_name: Name of the model to use
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        :return: dictionary of output files by the execution identifier
        """
        app_files: list[File] = await self._app.get_files()
        model_file: File = await check_and_get_model(self._app, model_name)
        model_metadata = get_model_metadata(model_file)

        files_by_run_id: list[dict] = []
        await asyncio.gather(
            *[
                check_and_store_output_file(
                    file,
                    files_by_run_id,
                    model_metadata=model_metadata,
                    file_name=file_name,
                    get_objects=get_objects,
                )
                for file in app_files
            ]
        )
        return files_by_run_id

    async def get_output_files_by_ai_function(
        self,
        ai_function: str,
        version: Optional[str] = None,
        file_name: Optional[str] = None,
        get_objects: bool = False,
    ):
        """Get output files for a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :return: The output files if they exist
        :param file_name: Name of the file to get
        :param get_objects: Get the file objects instead of the file metadata
        """
        app_files: list[File] = await self._app.get_files()
        activity_template: ActivityTemplate = await get_activity_template(
            self._universe, ai_function, version
        )

        files_by_run_id: list[dict] = []
        await asyncio.gather(
            *[
                check_and_store_output_file(
                    file,
                    files_by_run_id,
                    activity_template=activity_template,
                    file_name=file_name,
                    get_objects=get_objects,
                )
                for file in app_files
            ]
        )
        return files_by_run_id

    async def get_output_file_objects(
        self, run_id: str, file_name: Optional[str] = None
    ) -> Union[tuple[bytes, dict], dict[str, tuple[bytes, dict]]]:
        """Get an output file object for a given ai_function and run id
        :param run_id: ID of the executed run
        :param file_name: Name of the file to get
        """
        app_files: list[File] = await self._app.get_files()
        files_by_run_id: list[dict] = []
        await asyncio.gather(
            *[
                check_and_store_output_file(
                    file, files_by_run_id, run_id, file_name=file_name, get_objects=True
                )
                for file in app_files
            ]
        )
        if len(files_by_run_id) == 0:
            log_error(
                logger, f"No output files found for run ({run_id})", AIFunctionError
            )

        output_files_by_name = files_by_run_id[0]["output_files"]

        if file_name is None:
            return_dict = {}
            for file_name, file_dict in output_files_by_name.items():
                return_dict[file_name] = file_dict["object"], {
                    k: v for k, v in file_dict.items() if k != "object"
                }
            return return_dict

        if file_name not in output_files_by_name:
            log_error(
                logger,
                f"No output file ({file_name}) found for run ({run_id})",
                AIFunctionError,
            )

        file_dict = output_files_by_name[file_name]
        return file_dict["object"], {
            k: v for k, v in file_dict.items() if k != "object"
        }

    async def _create_input_file(
        self,
        file_name: str,
        file: Union[any, tuple[any, dict]],
        force_overwrite: bool = False,
    ):
        """Create an output file of an ai_function
        :param file_name: Name of the file to create
        :param file: File to be uploaded
        :param force_overwrite: Overwrite the file if it already exists
        """
        metadata = {}
        if isinstance(file, tuple):
            file, metadata = file

        complete_file_name = get_input_file_name(file_name)
        if (
            not force_overwrite
            and await self._app.get_file(name=complete_file_name) is not None
        ):
            log_error(
                logger,
                f"Input file ({file_name}) already exists, if you want to overwrite it, use force_overwrite=True",
                AIFunctionError,
            )
        await self._app.create_file(
            name=complete_file_name,
            file_object=file,
            tags=["shimoku_generated", "ai_input_file"],
            metadata={"file_name": file_name, **metadata},
        )
        logger.info(f"Created input file {file_name}")

    async def _get_input_file(self, file_name: str) -> str:
        """Get the ID of an input file
        :param file_name: Name of the file to get
        :return: ID of the file
        """
        file_name = get_input_file_name(file_name)
        input_file: Optional[File] = await self._app.get_file(name=file_name)

        if input_file is None:
            log_error(logger, f"Input file ({file_name}) not found", AIFunctionError)

        if (
            input_file["tags"] != ["shimoku_generated", "ai_input_file"]
            or "file_name" not in input_file["metadata"]
        ):
            log_error(
                logger, f"File ({file_name}) is not an input file", AIFunctionError
            )

        return input_file["id"]

    async def create_input_files(
        self, input_files: dict, force_overwrite: bool = False
    ):
        """Create input files for a ai_function
        :param input_files: dictionary of input files to create
        :param force_overwrite: Force the overwrite of the files
        """
        check_app_is_set(self)
        await asyncio.gather(
            *[
                self._create_input_file(file_name, file, force_overwrite)
                for file_name, file in input_files.items()
            ]
        )

    async def _get_available_input_files(self) -> list[File]:
        """Get available input files for a ai_function"""
        app_files: list[File] = await self._app.get_files()
        input_files = [
            file
            for file in app_files
            if "shimoku_generated" in file["tags"] and "ai_input_file" in file["tags"]
        ]
        return sorted(input_files, key=lambda x: x["metadata"]["file_name"])

    async def get_available_input_files(self) -> list[dict]:
        """Get available input files for a ai_function"""
        check_app_is_set(self)
        return [file["metadata"] for file in await self._get_available_input_files()]

    async def delete_input_file(self, file_name: str):
        """Delete an input file
        :param file_name: Name of the file to delete
        """
        check_app_is_set(self)
        file_id = await self._get_input_file(file_name)
        await self._app.delete_file(file_id)
        logger.info(f"Deleted input file {file_name}")

    async def generic_execute(
        self,
        ai_function: str,
        version: Optional[str] = None,
        universe_api_key: Optional[str] = None,
        **params,
    ) -> str:
        """Execute a generic ai_function
        :param ai_function: Name of the ai_function to execute
        :param version: Version of the ai_function to execute
        :param universe_api_key: API key of the universe
        :param params: Parameters to be passed to the ai_function
        """

        async def check_params(_activity_template: ActivityTemplate, _params: dict):
            """Check the parameters passed to the ai_function, and create input files if necessary
            :param _activity_template: Activity template of the ai_function
            :param _params: Parameters to be passed to the ai_function
            """
            input_settings = _activity_template["inputSettings"]
            if any(param not in input_settings for param in _params):
                log_error(
                    logger,
                    f"Unknown parameters for AI function ({_activity_template['name']}): "
                    f"{[param for param in _params if param not in input_settings]} \n"
                    f"The possible parameters are: {list(input_settings.keys())}",
                    AIFunctionError,
                )
            for param_name, definition in input_settings.items():
                if param_name not in _params:
                    if definition["mandatory"]:
                        log_error(
                            logger,
                            f"Missing parameter ({param_name}) for AI function ({_activity_template['name']}), "
                            f"the description of the missing parameter is: {definition['description']}",
                            AIFunctionError,
                        )
                    else:
                        continue
                param_value = _params[param_name]
                param_definition_type = definition["datatype"]
                if definition["datatype"] == "file":
                    if not isinstance(param_value, str):
                        log_error(
                            logger,
                            f"The parameter ({param_name}) is expected to be a reference name to an input file",
                            AIFunctionError,
                        )
                    _params[param_name] = await self._get_input_file(param_value)
                elif definition["datatype"] == "model":
                    if not isinstance(param_value, str):
                        log_error(
                            logger,
                            f"The parameter ({param_name}) is expected to be a model name",
                            AIFunctionError,
                        )
                    if not await check_and_get_model(self._app, param_value):
                        log_error(
                            logger,
                            f"The model ({param_value}) does not exist",
                            AIFunctionError,
                        )
                elif str(type(param_value)) != f"<class '{param_definition_type}'>":
                    log_error(
                        logger,
                        f"Wrong type for parameter ({param_name}) for the AI function ({_activity_template['name']})"
                        f"the description of the missing parameter is: {definition['description']}\n"
                        f"Expected type: {param_definition_type}\n"
                        f"Provided type: {str(type(param_value))[8:-2]}",
                        AIFunctionError,
                    )

        activity_template: ActivityTemplate = await get_activity_template(
            self._universe, ai_function, version
        )
        activity: Activity = await get_activity_from_template(
            self._app, activity_template, universe_api_key
        )
        await check_params(activity_template, params)
        run: Activity.Run = await activity.create_run(settings=params)
        logger.info(f"Response from execution trigger: {await run.trigger_webhook()}")
        return run["id"]



from typing import Union, Optional, TYPE_CHECKING

import pandas as pd

from shimoku.exceptions import BentoboxError
from shimoku.plt.utils import create_normalized_name, ShimokuPalette
from shimoku_components_catalog.html_components import create_h1_title_with_modal

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer

import logging
from shimoku.execution_logger import log_error

logger = logging.getLogger(__name__)


def check_for_bentobox(self: "PlotLayer"):
    if self._bentobox_data:
        log_error(
            logger,
            "This function groups a chart and a group of indicators with a bentobox,"
            " so it cannot be used inside another bentobox. Please"
            " pop out of the current bentobox before using this function.",
            BentoboxError,
        )


def _call_inner_chart_function_with_parameters(
    self: "PlotLayer",
    order: int,
    default_rows_size: int,
    default_cols_size: int,
    chart_parameters: dict,
    chart_function: Optional[callable] = None,
) -> any:
    if not chart_parameters or not isinstance(chart_parameters, dict):
        chart_parameters = {}

    restricted_parameters = ["order"]

    for parameter in restricted_parameters:
        if parameter in chart_parameters:
            log_error(
                logger,
                f"Parameter {parameter} cannot be used in the inner chart",
                ValueError,
            )

    chart_parameters.update(
        dict(
            order=order,
            padding="1,0,0,1"
            if "padding" not in chart_parameters
            else chart_parameters["padding"],
            rows_size=default_rows_size
            if "rows_size" not in chart_parameters
            else chart_parameters["rows_size"],
            cols_size=default_cols_size
            if "cols_size" not in chart_parameters
            else chart_parameters["cols_size"],
        )
    )

    if chart_function is None:
        chart_function = self.bar

    return chart_function(**chart_parameters)


def infographics_text_bubble(
    self: "PlotLayer",
    title: str,
    text: str,
    order: int,
    chart_parameters: dict,
    rows_size: int = 3,
    cols_size: int = 12,
    chart_function: Optional[callable] = None,
    background_url: Optional[str] = None,
    background_color: str = ShimokuPalette.BACKGROUND.value,
    bubble_location: str = "top",
    image_url: Optional[str] = None,
    image_size: int = 100,
):
    if bubble_location not in ["top", "bottom", "left", "right"]:
        log_error(logger, f"Invalid location value '{bubble_location}'.", ValueError)

    chart_first = bubble_location in ["bottom", "right"]
    vertical = bubble_location in ["top", "bottom"]

    if image_url and vertical:
        log_error(logger, "The image cannot be displayed vertically.", ValueError)

    default_rows_size = (
        int(10 * rows_size * (2 / 3)) - 2 * int(bubble_location == "bottom")
        if vertical
        else 10 * rows_size
    )
    default_cols_size = 23 if vertical else 15

    if image_url == "default":
        image_url = "https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63332131120af18c03d2b69a_APAME-about.svg"

    check_for_bentobox(self)
    self.set_bentobox(cols_size, rows_size)

    _call_inner_chart_function_with_parameters(
        self=self,
        order=order + int(not chart_first),
        default_rows_size=default_rows_size,
        default_cols_size=default_cols_size,
        chart_function=chart_function,
        chart_parameters=chart_parameters,
    )

    chart_padding = chart_parameters["padding"].replace(" ", "").split(",")
    sides_padding = int(chart_padding[1]) + int(chart_padding[3])
    vertical_padding = int(chart_padding[0]) + int(chart_padding[2])

    html_cols_size = 22 - int(not vertical) * (
        chart_parameters["cols_size"] + sides_padding
    )
    html_rows_size = 10 * rows_size - int(vertical) * (
        chart_parameters["rows_size"] + vertical_padding
    )

    r_hash = create_normalized_name(self._get_component_hash(order))

    html = (
        "<head>"
        f"<style>.block-text-{r_hash}{{padding:24px; "
        f"color: {ShimokuPalette.BLACK.value}; font-size: 16px;"
        + (f"background-color: {background_color}; " if background_url is None else "")
        + f"background-size:cover; background-repeat: no-repeat;background-image: url('{background_url}');"
        "background-position: center; border-radius:var(--border-radius-m);}</style>"
        "</head>"
        f"<div class='block-text-{r_hash}'>"
        + (f"<img src={image_url} width='{image_size}%'>" if image_url else "")
        + f"<h1>{title}</h1>"
        f"<p>{text}</p>"
        "</div>"
    )
    self.html(
        html=html,
        order=order + int(chart_first),
        rows_size=html_rows_size,
        cols_size=html_cols_size,
        padding=f'{int(bubble_location != "bottom")},1,1,1',
    )
    self.pop_out_of_bentobox()


def chart_and_modal_button(
    self: "PlotLayer",
    order: int,
    chart_parameters: dict,
    button_modal: str,
    rows_size: int = 3,
    cols_size: int = 12,
    chart_function: Optional[callable] = None,
    button_label: str = "Read more",
    button_side_text: str = "Click on the button to read more about this topic.",
):
    check_for_bentobox(self)
    self.set_bentobox(cols_size, rows_size)
    _call_inner_chart_function_with_parameters(
        self=self,
        order=order,
        default_rows_size=rows_size * 10 - 6,
        default_cols_size=22,
        chart_function=chart_function,
        chart_parameters=chart_parameters,
    )

    chart_padding = chart_parameters["padding"].replace(" ", "").split(",")
    vertical_padding = int(chart_padding[0]) + int(chart_padding[2])

    button_rows_size = (
        10 * rows_size - chart_parameters["rows_size"] + vertical_padding - 2
    )

    if cols_size > 3:
        html = (
            "<head>"
            f"<style>.rubik-text{{padding: 2%;"
            f"background-color: {ShimokuPalette.STRIPE_LIGHT.value};"
            f"background-size:cover; background-repeat: no-repeat;"
            "background-position: center; border-radius:var(--border-radius-m);"
            f"color: {ShimokuPalette.GRAY.value}; font-size: 16px;}}</style>"
            "</head>"
            "<div class='rubik-text'>"
            f"<p>{button_side_text}</p>"
            "</div>"
        )

        self.html(
            html=html,
            order=order + 1,
            rows_size=button_rows_size,
            cols_size=13 + int(0.5 * (cols_size - 4)),
            padding="0,1,0,1",
        )

    self.modal_button(
        order=order + 2,
        modal=button_modal,
        label=button_label,
        rows_size=button_rows_size,
        cols_size=2,
        padding="0,0,0,1",
    )
    self.pop_out_of_bentobox()


def chart_and_indicators(
    self: "PlotLayer",
    order: int,
    chart_parameters: dict,
    indicators_groups: list[Union[pd.DataFrame, list[dict]]],
    indicators_parameters: Optional[dict] = None,
    chart_rows_size: int = 3,
    cols_size: int = 12,
    chart_function: Optional[callable] = None,
) -> int:
    check_for_bentobox(self)
    if indicators_parameters is None:
        indicators_parameters = {}

    self.set_bentobox(cols_size, chart_rows_size + len(indicators_groups))

    _call_inner_chart_function_with_parameters(
        self=self,
        order=order,
        default_rows_size=chart_rows_size * 14,
        default_cols_size=22,
        chart_function=chart_function,
        chart_parameters=chart_parameters,
    )

    not_allowed_parameters = ["data", "menu_path", "order", "bentobox_data"]
    for parameter in not_allowed_parameters:
        if parameter in indicators_parameters:
            log_error(
                logger,
                f"Parameter {parameter} cannot be used in the inner indicators",
                ValueError,
            )

    indicators_parameters["padding"] = indicators_parameters.get("padding", "0,0,0,1")
    indicators_parameters["rows_size"] = indicators_parameters.get("rows_size", 8)
    indicators_parameters["cols_size"] = indicators_parameters.get("cols_size", 23)

    order += 1
    for indicators_group in indicators_groups:
        order = self.indicator(
            data=indicators_group, order=order, **indicators_parameters
        )

    self.pop_out_of_bentobox()

    return order


def indicators_with_header(
    self: "PlotLayer",
    order: int,
    indicators_groups: list[Union[pd.DataFrame, list[dict]]],
    title: str,
    indicators_parameters: Optional[dict] = None,
    subtitle: str = "",
    background_color: str = ShimokuPalette.PRIMARY.value,
    text_color: str = ShimokuPalette.BACKGROUND_PAPER.value,
    cols_size: int = 12,
    icon_url: str = "https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/63e3615716d4435d29e0b82c_Acurracy.svg",
) -> int:
    if indicators_parameters is None:
        indicators_parameters = {}
    return chart_and_indicators(
        self=self,
        order=order,
        chart_rows_size=1,
        cols_size=cols_size,
        chart_function=self.html,
        chart_parameters=dict(
            html=create_h1_title_with_modal(
                title=title,
                subtitle=subtitle,
                background_color=background_color,
                text_color=text_color,
                icon_url=icon_url,
            ),
            padding="0,0,0,0",
            cols_size=24,
        ),
        indicators_groups=indicators_groups,
        indicators_parameters=indicators_parameters,
    )


def line_with_summary(
    self: "PlotLayer",
    order: int,
    data: Union[str, pd.DataFrame, list[dict]],
    title: str,
    x: str,
    y: Optional[Union[str, list[str]]] = None,
    description: str = "",
    value: Union[str, float, int] = "",
    rows_size: int = 2,
    cols_size: int = 2,
):
    check_for_bentobox(self)

    self.set_bentobox(cols_size, rows_size)

    _call_inner_chart_function_with_parameters(
        self=self,
        order=order,
        default_rows_size=(rows_size // 2 + rows_size % 2) * 12,
        default_cols_size=22,
        chart_function=self.area,
        chart_parameters={
            "data": data,
            "x": x,
            "y": y,
            "variant": "minimal",
            "option_modifications": {
                "xAxis": {
                    "axisLabel": {
                        "show": True,
                        "showMaxLabel": True,
                        "showMinLabel": True,
                        "interval": 10000000000000,
                    },
                    "boundaryGap": True,
                },
                "yAxis": {
                    "axisLabel": {"show": False},
                },
                "grid": {"left": "1%", "right": "1%", "top": "1%", "bottom": "11%"},
            },
        },
    )
    self.indicator(
        data={
            "title": title,
            "description": description,
            "value": value,
            "align": "center",
            "color": "var(--color-black)",
        },
        order=order + 1,
        rows_size=(rows_size // 2 + rows_size % 2) * 8,
        cols_size=22,
        padding="0,0,0,1",
    )
    self.pop_out_of_bentobox()


def table_with_header(
    self: "PlotLayer",
    order: int,
    data: Union[str, pd.DataFrame, list[dict]],
    title: str,
    table_parameters: Optional[dict] = None,
    subtitle: str = "",
    rows_size: int = 3,
    cols_size: int = 4,
    modal_title: Optional[str] = None,
    icon_url: str = "https://uploads-ssl.webflow.com/619f9fe98661d321dc3beec7/"
    "63594ccf3f311a98d72faff7_suite-customer-b.svg",
    modal_text: str = "You can click the X in the corner or click the overlay to close this modal. "
    "This is a nice way to show additional information.",
    text_color: str = ShimokuPalette.BACKGROUND_PAPER.value,
    background_color: str = ShimokuPalette.BLACK.value,
    modal_icon_color: str = ShimokuPalette.CHART_C1.value,
    modal_icon_hover_color: str = ShimokuPalette.PRIMARY_DARK.value,
):
    check_for_bentobox(self)
    self.set_bentobox(cols_size=cols_size, rows_size=rows_size)
    self.html(
        html=create_h1_title_with_modal(
            title=title,
            subtitle=subtitle,
            background_color=background_color,
            text_color=text_color,
            icon_url=icon_url,
            modal_title=modal_title,
            modal_text=modal_text,
            modal_icon_color=modal_icon_color,
            modal_icon_hover_color=modal_icon_hover_color,
        ),
        order=order,
        cols_size=24,
        rows_size=1,
        padding="0,0,0,0",
    )

    if table_parameters is None:
        table_parameters = {}
    table_parameters["data"] = data
    table_parameters["padding"] = table_parameters.get("padding", "0,0,0,1")

    _call_inner_chart_function_with_parameters(
        self=self,
        order=order + 1,
        default_rows_size=rows_size,
        default_cols_size=22,
        chart_function=self.table,
        chart_parameters=table_parameters,
    )
    self.pop_out_of_bentobox()



sunburst_data = [
    {
        "name": "Root 1",
        "children": [
            {
                "name": "Children A",
                "value": 15,
                "children": [
                    {"name": "Children A1", "value": 2},
                    {
                        "name": "Children AA1",
                        "value": 5,
                        "children": [{"name": "Children AAA1", "value": 2}],
                    },
                    {"name": "Children A2", "value": 4},
                ],
            },
            {
                "name": "Children B",
                "value": 10,
                "children": [
                    {"name": "Children B1", "value": 5},
                    {"name": "Children B2", "value": 1},
                ],
            },
        ],
    },
    {
        "name": "Root 2",
        "children": [
            {
                "name": "Children A1",
                "children": [
                    {"name": "Children AA1", "value": 1},
                    {"name": "Children AA2", "value": 2},
                ],
            }
        ],
    },
]



from typing import Optional, Union, TYPE_CHECKING

from shimoku.plt.EChart_definitions.default_echart_options import (
    get_common_echart_options,
    get_common_series_options,
)

if TYPE_CHECKING:
    from shimoku.plt.plt_layer import PlotLayer


async def sunburst_chart(
    self: "PlotLayer",
    order: int,
    data: Union[str, dict, list],
    title: Optional[str] = None,
    padding: Optional[str] = None,
    rows_size: Optional[int] = None,
    cols_size: Optional[int] = None,
    option_modifications: Optional[dict] = None,
):
    """Create a sunburst chart"""
    common_options = get_common_echart_options()
    series_options = get_common_series_options()
    del common_options["xAxis"]
    del common_options["yAxis"]
    del common_options["toolbox"]["feature"]["magicType"]
    if isinstance(data, dict):
        data = [data]
    series_options.update(
        {
            "type": "sunburst",
            "data": "#set_data#",
            "radius": [0, "95%"],
            "label": {
                "rotate": "radial",
            },
        }
    )
    common_options.update(
        {
            "tooltip": {"trigger": "item", "triggerOn": "mousemove"},
            "series": series_options,
        }
    )
    await self._create_echart(
        data_mapping_to_tuples=await self._choose_data(order, data, dump_whole=True),
        fields=["data"],
        options=common_options,
        order=order,
        title=title,
        rows_size=rows_size,
        cols_size=cols_size,
        padding=padding,
        option_modifications=option_modifications,
    )



from typing import Optional
from shimoku.api.resources.universe import Universe
from shimoku import ApiClient
from shimoku.exceptions import UniverseError

import logging
from shimoku.execution_logger import ClassWithLogging, log_error

logger = logging.getLogger(__name__)


class UniversesLayer(ClassWithLogging):
    """
    Class used to interact with the API at the universe level
    """

    _module_logger = logger
    _use_info_logging = True

    def __init__(self, api_client: ApiClient):
        self._api_client = api_client

    async def create_universe_api_key(self, uuid: str, description: str) -> dict:
        """
        Create a universe API key, this key has admin privileges.
        :param uuid: uuid of the universe
        :param description: description of the key
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return (await universe.create_universe_api_key(description)).cascade_to_dict()

    async def get_universe_api_keys(self, uuid: str) -> list[dict]:
        """
        Get the universe API keys.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [
            universe_api_key.cascade_to_dict()
            for universe_api_key in await universe.get_universe_api_keys()
        ]

    async def delete_universe_api_key(self, uuid: str, api_key_uuid: str) -> bool:
        """
        Delete a universe API key.
        :param uuid: uuid of the universe
        :param api_key_uuid: uuid of the API key
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return await universe.delete_universe_api_key(api_key_uuid)

    async def get_universe_workspaces(self, uuid: str) -> list[dict]:
        """
        Get the universe workspaces.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [b.cascade_to_dict() for b in await universe.get_businesses()]

    async def get_universe_activity_templates(self, uuid: str) -> list[dict]:
        """
        Get the universe activity templates as a list of dictionaries.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [
            {
                "name": at["name"],
                "description": at["description"],
                "min_run_interval": at["minRunInterval"],
                "input_settings": [
                    {
                        "name": name,
                        "description": param["description"],
                        "datatype": param["datatype"],
                    }
                    for name, param in at["inputSettings"].items()
                    if isinstance(param, dict)
                ],
            }
            for at in await universe.get_activity_templates()
            if at["enabled"]
        ]

    async def get_universe_actions(self, uuid: Optional[str] = None) -> list[dict]:
        """
        Get the universe actions_execution as a list of dictionaries.
        :param uuid: uuid of the universe
        """
        if self._api_client.playground:
            uuid = "local"
        elif not uuid:
            log_error(
                logger,
                "If not in playground, a universe uuid is required",
                UniverseError,
            )
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [action.cascade_to_dict() for action in await universe.get_actions()]

    async def get_universe_action_templates(self, uuid: str) -> list[dict]:
        """
        Get the universe action templates as a list of dictionaries.
        :param uuid: uuid of the universe
        """
        universe = Universe(api_client=self._api_client, uuid=uuid)
        return [
            {
                "name": action_template["name"],
                "description": action_template["description"],
                "public": action_template["public"],
                "libraries": action_template["pythonLibraries"],
            }
            for action_template in await universe.get_action_templates()
        ]



from ...report import Report


class AnnotatedEChart(Report):
    report_type = "ANNOTATED_ECHART"

    default_properties = dict(
        **Report.default_properties,
        option={},
        slider={},
    )



from shimoku import Client


def gauge_indicators(shimoku_client: Client):
    shimoku_client.set_menu_path("test-free-echarts", "gauge-indicator")

    shimoku_client.plt.gauge_indicator(
        order=0,
        value=83,
        description="Síntomas coincidientes | Mareo, Dolor cervical",
        title="Sobrecarga muscular en cervicales y espalda",
    )

    shimoku_client.plt.gauge_indicator(
        order=2,
        value=31,
        color=2,
        description="Síntomas coincidientes | Dolor cervical",
        title="Bruxismo",
    )
